import math


def get_start_and_row_count(limit=None, offset=0, page_size=50):
    """Many API's have a start and row count parameter. This function
    converts an offset and limit into a iterator of start and row count
    tuples. If limit is None, then the iterator will be infinite.
    Otherwise, the iterator will be finite and the last row count may
    be less than the page size."""
    offset = offset or 0
    if not limit:
        page_no = 0
        while True:
            yield (page_no * page_size + offset, page_size)
            page_no += 1
    else:
        num_full_pages = math.floor(limit / page_size)
        last_page_size = limit % page_size
        for i in range(num_full_pages):
            yield (i * page_size + offset, page_size)
        if last_page_size:
            yield (num_full_pages * page_size + offset, last_page_size)



from typing import Optional

from pydantic import Field

from .schema import NumberServiceResultSchema
from patent_client.epo.ops.util import EpoBaseModel
from patent_client.util.pydantic_util import Date


class DocumentId(EpoBaseModel):
    doc_type: Optional[str] = None
    id_type: Optional[str] = None
    country: Optional[str] = None
    number: Optional[str] = None
    kind: Optional[str] = None
    date: Optional[Date] = None
    name: Optional[str] = None

    def __str__(self):
        return f"{self.country}{self.number}{self.kind}"


class NumberServiceResult(EpoBaseModel):
    __schema__ = NumberServiceResultSchema()
    input_doc: DocumentId
    output_doc: DocumentId
    service_version: Optional[str] = None
    messages: list = Field(default_factory=list)



import logging
from io import BytesIO

from yankee.data import AttrDict

from .model.biblio import BiblioResult
from .model.fulltext import Claims
from .model.fulltext import Description
from .model.images import Images
from .model.search import Search
from patent_client.epo.ops.session import asession

logger = logging.getLogger(__name__)


class PublishedBiblioAsyncApi:
    @classmethod
    async def get_constituents(cls, number, doc_type="publication", format="docdb", constituents=("biblio",)):
        """Published Data Constituents API
        number: document number to search
        doc_type: document type (application / publication)
        format: document number format (original / docdb / epodoc)
        constituents: what data to retrieve. Can be combined. (biblio / abstract / full-cycle)

        """
        base_url = f"http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/{format}/{number}/"
        if isinstance(constituents, str):
            constituents = (constituents,)
        url = base_url + ",".join(constituents)
        response = await asession.get(url)

        return response.text

    @classmethod
    async def get_biblio(cls, number, doc_type="publication", format="docdb") -> "BiblioResult":
        text = await cls.get_constituents(number, doc_type, format, constituents="biblio")
        return BiblioResult.model_validate(text)

    @classmethod
    async def get_abstract(cls, number, doc_type="publication", format="docdb"):
        return await cls.get_constituents(number, doc_type, format, constituents="abstract")

    @classmethod
    async def get_full_cycle(cls, number, doc_type="publication", format="docdb"):
        return await cls.get_constituents(number, doc_type, format, constituents="full-cycle")


class PublishedFulltextAsyncApi:
    fulltext_jurisdictions = "EP, WO, AT, BE, BG, CA, CH, CY, CZ, DK, EE, ES, FR, GB, GR, HR, IE, IT, LT, LU, MC, MD, ME, NO, PL, PT, RO, RS, SE, SK".split(
        ", "
    )

    @classmethod
    async def get_fulltext_result(cls, number, doc_type="publication", format="docdb", inquiry="fulltext"):
        """Published Fulltext API
        number: document number to search
        doc_type: document type (application / publication)
        format: document number format (original / docdb / epodoc)
        inquiry: what data to retrieve. Can be combined. (fulltext / description / claims)

        """
        url = f"http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/{format}/{number}/{inquiry}"
        if number[:2] not in cls.fulltext_jurisdictions:
            raise ValueError(
                f"Fulltext Is Not Available For Country Code {number[:2]}. Fulltext is only available in {', '.join(cls.fulltext_jurisdictions)}"
            )
        response = await asession.get(url)
        response.raise_for_status
        return response.text

    @classmethod
    async def get_description(cls, number, doc_type="publication", format="docdb"):
        text = await cls.get_fulltext_result(number, doc_type="publication", format="docdb", inquiry="description")
        return Description.model_validate(text)

    @classmethod
    async def get_claims(cls, number, doc_type="publication", format="docdb") -> "Claims":
        text = await cls.get_fulltext_result(number, doc_type="publication", format="docdb", inquiry="claims")

        return Claims.model_validate(text)


class PublishedSearchAsyncApi:
    @classmethod
    async def search(cls, query, start=1, end=100):
        base_url = "http://ops.epo.org/3.2/rest-services/published-data/search"
        range = f"{start}-{end}"
        logger.debug(f"OPS Search Endpoint - Query: {query}\nRange: {start}-{end}")
        response = await asession.get(base_url, params={"Range": range, "q": query})
        if response.status_code == 404:
            return AttrDict.convert(
                {
                    "query": "query",
                    "num_results": 0,
                    "begin": start,
                    "end": end,
                    "results": list(),
                }
            )

        return Search.model_validate(response.text)


class PublishedImagesAsyncApi:
    @classmethod
    async def get_images(cls, number, doc_type="publication", format="docdb"):
        base_url = f"http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/{format}/{number}/images"
        response = await asession.get(base_url)

        return Images.model_validate(response.text)

    @classmethod
    async def get_page_image(cls, country, number, kind, image_type, page_number, image_format="pdf"):
        response = await asession.get(
            f"https://ops.epo.org/3.2/rest-services/published-data/images/{country}/{number}/{kind}/{image_type}.{image_format}",
            params={"Range": page_number},
            stream=True,
        )

        return BytesIO(response.raw.read())

    @classmethod
    async def get_page_image_from_link(cls, link, page_number, image_format="pdf"):
        response = await asession.get(
            f"https://ops.epo.org/3.2/rest-services/{link}.{image_format}",
            params={"Range": page_number},
            stream=True,
        )

        return BytesIO(response.raw.read())


class PublishedAsyncApi:
    biblio = PublishedBiblioAsyncApi
    fulltext = PublishedFulltextAsyncApi
    search = PublishedSearchAsyncApi
    images = PublishedImagesAsyncApi



import logging
from collections.abc import Sequence
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import AsyncIterator
from typing import TYPE_CHECKING

from pypdf import PdfMerger

from .api import PatentExaminationDataSystemApi
from .query import QueryFields
from .util import date_to_solr_date
from patent_client.util.asyncio_util import run_sync
from patent_client.util.manager import Manager
from patent_client.util.request_util import get_start_and_row_count

if TYPE_CHECKING:
    from .model import USApplication, Document

logger = logging.getLogger(__name__)


class HttpException(Exception):
    pass


class USApplicationManager(Manager["USApplication"]):
    default_filter = "appl_id"

    async def alen(self):
        api = PatentExaminationDataSystemApi()
        max_length = (await api.create_query(**self.get_query_params())).num_found
        return min(max_length, self.config.limit) if self.config.limit else max_length

    async def _aget_results(self) -> AsyncIterator["USApplication"]:
        query_params = self.get_query_params()
        api = PatentExaminationDataSystemApi()
        for start, rows in get_start_and_row_count(self.config.limit, self.config.offset, page_size=20):
            page = await api.create_query(**{**query_params, "start": start, "rows": rows})
            for app in page.applications:
                yield app
            if len(page.applications) < rows:
                break

    def get_query_params(self):
        # Short circuit processing logic if the "query" filter is specified
        if "query" in self.config.filter:
            query_text = self.config.filter["query"]
        else:
            query = list()
            date_filters = {
                QueryFields.get(k): v for k, v in self.config.filter.items() if QueryFields.is_date_field(k)
            }
            non_date_filters = {QueryFields.get(k): v for k, v in self.config.filter.items() if k not in date_filters}

            date_filter_tuples = list()
            # Check date filters for validity
            for k, v in date_filters.items():
                if "__" in k:
                    f, *op = k.split("__")
                    if len(op) > 1:
                        raise ValueError(f"Invalid date filter: {k}={v}; Cannot have more than one operator")
                    if op[0] not in ("gte", "lte"):
                        raise ValueError(f"Invalid date filter: {k}={v}; Invalid operator: {op[0]}")
                    if isinstance(v, (list, tuple)):
                        raise ValueError(
                            f"Invalid date filter: {k}={v}; Cannot have multiple values with operator {op[0]}"
                        )
                    date_filter_tuples.append((f, op[0], date_to_solr_date(v)))
                else:
                    if isinstance(v, (list, tuple)):
                        if len(v) != 2:
                            raise ValueError(
                                f"Invalid date range filter: {k}; When filtering using a list or tuple, must have length 2 (start, end)"
                            )
                        date_filter_tuples.append((k, "gte", date_to_solr_date(v[0])))
                        date_filter_tuples.append((k, "lte", date_to_solr_date(v[1])))
                    else:
                        date_filter_tuples.append((k, "exact", date_to_solr_date(v)))

            # Create pairs of gte/lte filters
            query_date_filter_tuples = set()
            for k, op, v in date_filter_tuples:
                if op == "gte":
                    lte_query = next((v for k, op, v in date_filter_tuples if k == k and op == "lte"), "*")
                    query_date_filter_tuples.append((k, (v, lte_query)))
                elif op == "lte":
                    gte_query = next((v for k, op, v in date_filter_tuples if k == k and op == "gte"), "*")
                    query_date_filter_tuples.append((k, (gte_query, v)))
                elif op == "exact":
                    query_date_filter_tuples.append((k, v))

            # Create the query string
            for k, v in query_date_filter_tuples:
                if isinstance(v, (list, tuple)):
                    query.append(f"{k}:[{v[0]} TO {v[1]}]")
                else:
                    query.append(f"{k}:{v}")

            # Add non-date filters
            for k, v in non_date_filters.items():
                if isinstance(v, Sequence) and not isinstance(v, str):
                    values = " OR ".join(str(i) for i in v)
                    query.append(f"{k}:({values})")
                else:
                    query.append(f"{k}:({v})")
            query_text = " AND ".join(query)

        if self.config.order_by:
            sort_query = ""
            for s in self.config.order_by:
                if s[0] == "-":
                    sort_query += f"{QueryFields.get(s)} desc ".strip()
                else:
                    sort_query += (f"{QueryFields.get(s)} asc").strip()
        else:
            sort_query = None

        mm = "0%" if "appEarlyPubNumber" not in query else "90%"

        query_data = {
            "query": query_text,
            "facet": False,
            "minimum_match": mm,
        }
        if sort_query:
            query_data["sort"] = sort_query
        return query_data

    @property
    def allowed_filters(self):
        return QueryFields.field_names()

    async def ais_online(self):
        return await PatentExaminationDataSystemApi.is_online()

    def is_online(self):
        return run_sync(self.ais_online())


class DocumentManager(Manager["Document"]):
    default_filter = "appl_id"

    async def alen(self):
        return len([i async for i in self])

    async def _aget_results(self) -> AsyncIterator["Document"]:
        api = PatentExaminationDataSystemApi()
        docs = await api.get_documents(self.config.filter["appl_id"])
        for doc in docs:
            yield doc

    def download(self, docs, path="."):
        run_sync(self.adownload(docs, path))

    async def adownload(self, docs, path="."):
        if str(path)[-4:].lower() == ".pdf":
            # If we've been given a specific filename, use it
            out_file = Path(path)
        else:
            out_file = Path(path) / "package.pdf"

        files = list()
        try:
            with TemporaryDirectory() as tmpdir:
                for doc in docs:
                    if doc.access_level_category == "PUBLIC":
                        files.append((await doc.adownload(tmpdir), doc))

                out_pdf = PdfMerger()
                page = 0
                for f, doc in files:
                    bookmark = f"{doc.mail_room_date} - {doc.code} - {doc.description}"
                    out_pdf.append(str(f), bookmark=bookmark, import_bookmarks=False)
                    page += doc.page_count

                out_pdf.write(str(out_file))
        except (PermissionError, NotADirectoryError):
            # This is needed due to a bug in Windows that prevents cleanup of the tmpdir
            pass
        return out_file



from typing import Optional

from pydantic import Field

from .schema import LegalSchema
from patent_client.epo.ops.number_service.model import DocumentId
from patent_client.epo.ops.util import EpoBaseModel
from patent_client.util.pydantic_util import Date


class MetaData(EpoBaseModel):
    status_of_data: Optional[str] = None
    docdb_publication_number: Optional[str] = None
    subscriber_exchange_date: Optional[Date] = None
    epo_created_date: Optional[Date] = None
    docdb_integer: Optional[int] = None


class LegalEvent(EpoBaseModel):
    """
    Field descriptions are here:
    https://documents.epo.org/projects/babylon/eponot.nsf/0/EF223017D933B30AC1257B50005A042E/$File/14.11_User_Documentation_3.1_en.pdf

    """

    filing_or_publication: Optional[str] = None
    document_number: Optional[str] = None
    ip_type: Optional[str] = None
    metadata: Optional[MetaData] = None
    country_code: Optional[str] = None
    text_record: Optional[str] = None

    event_date: Optional[Date] = None
    event_code: Optional[str] = None
    event_country: Optional[str] = None
    event_description: Optional[str] = None
    regional_event_code: Optional[str] = None

    corresponding_patent: Optional[str] = None
    corresponding_patent_publication_date: Optional[Date] = None

    designated_states: Optional[str] = None
    extension_name: Optional[str] = None
    new_owner_name: Optional[str] = None
    free_text: Optional[str] = None
    spc_number: Optional[str] = None
    spc_filing_date: Optional[str] = None
    expiry: Optional[str] = None
    publication_language: Optional[str] = None
    inventor_name: Optional[str] = None
    ipc_class: Optional[str] = None
    representative_name: Optional[str] = None
    payment_date: Optional[str] = None
    opponent_name: Optional[str] = None
    year_of_fee_payment: Optional[str] = None
    name_of_requester: Optional[str] = None
    date_extension_granted: Optional[str] = None
    extension_states: Optional[str] = None
    effective_date: Optional[Date] = None
    date_of_withdrawal: Optional[str] = None

    def __repr__(self):
        return f"Event(document_number={self.document_number}, event_description={self.event_description}, event_date={self.event_date})"


class Legal(EpoBaseModel):
    __schema__ = LegalSchema()
    publication_reference: Optional[DocumentId] = None
    events: list = Field(default_factory=list)



import json
from pathlib import Path

fixtures = Path(__file__).parent / "fixtures"

from .convert.biblio import PublicSearchBiblioPageSchema
from .convert.document import PublicSearchDocumentSchema


def test_convert_biblio():
    input_file = fixtures / "biblio.json"
    output_file = fixtures / "biblio_convert.json"
    input_data = json.loads(input_file.read_text())
    output_data = PublicSearchBiblioPageSchema().deserialize(input_data)
    # output_file.write_text(output_data.to_json(indent=2))
    expected_data = json.loads(output_file.read_text())
    assert json.loads(output_data.to_json()) == expected_data


def test_convert_doc():
    input_file = fixtures / "doc.json"
    output_file = fixtures / "doc_convert.json"
    input_data = json.loads(input_file.read_text())
    output_data = PublicSearchDocumentSchema().load_batch(input_data)
    output_file.write_text(output_data.to_json(indent=2))
    expected_data = json.loads(output_file.read_text())
    assert json.loads(output_data.to_json()) == expected_data



import datetime

from yankee.json.schema import fields as f
from yankee.json.schema import RegexSchema
from yankee.json.schema import Schema
from yankee.json.schema import ZipSchema

from ..util import html_to_text
from .shared import DocumentStructureSchema
from patent_client.util.claims.parser import ClaimsParser


def parse_claims(html):
    text = html_to_text(html)
    if text:
        try:
            return ClaimsParser().parse(text)
        except Exception:
            return list()
    return list()


class DocumentSchema(Schema):
    abstract_html = f.String("abstractHtml")
    government_interest = f.String("governmentInterest")
    background_html = f.String("backgroundTextHtml")
    brief_html = f.String("briefHtml")
    description_html = f.String("descriptionHtml")
    claim_statement = f.String("claimStatement")
    claims_html = f.String("claimsHtml")
    claims = f.String("claimsHtml", formatter=parse_claims)


class UsReferenceSchema(ZipSchema):
    publication_number = f.String("urpn")
    # us_class = f.String("usRefClassification")
    # cpc_class = f.String("usRefCpcClassification")
    # group = f.String("usRefGroup")
    pub_month = f.Date(
        "usRefIssueDate",
        dt_converter=lambda s: datetime.datetime(year=int(s[:4]), month=int(s[4:6]), day=1),
    )
    patentee_name = f.String("usRefPatenteeName")
    cited_by_examiner = f.Boolean("usRefGroup", true_func=lambda s: "examiner" in s)


class ForeignReferenceSchema(ZipSchema):
    citation_classification = f.String("foreignRefCitationClassification")
    citation_cpc = f.String("foreignRefCitationCpc")
    country_code = f.String("foreignRefCountryCode")
    # group = f.String("foreignRefGroup")
    patent_number = f.String("foreignRefPatentNumber")
    pub_month = f.Date(
        "foreignRefPubDate",
        dt_converter=lambda s: datetime.datetime(year=int(s[:4]), month=int(s[4:6]), day=1),
    )
    cited_by_examiner = f.Boolean("foreignRefGroup", true_func=lambda s: "examiner" in s)


class NplReferenceSchema(RegexSchema):
    __regex__ = r"(?P<citation>.*)(?P<cited_by_examiner>cited by (applicant|examiner).?$)"
    citation = f.String()
    cited_by_examiner = f.Bool(true_func=lambda s: "examiner" in s)


class RelatedApplicationSchema(ZipSchema):
    child_patent_country = f.String("relatedApplChildPatentCountry")
    child_patent_number = f.String("relatedApplChildPatentNumber")
    country_code = f.String("relatedApplCountryCode")
    filing_date = f.Date("relatedApplFilingDate")
    number = f.String("relatedApplNumber")
    parent_status_code = f.String("relatedApplParentStatusCode")
    patent_issue_date = f.Date("relatedApplPatentIssueDate")
    patent_number = f.String("relatedApplPatentNumber")


class InventorSchema(ZipSchema):
    name = f.String("inventorsName")
    city = f.String("inventorCity")
    country = f.String("inventorCountry")
    postal_code = f.String("inventorPostalCode")
    state = f.String("inventorState")


class ApplicantSchema(ZipSchema):
    city = f.String("applicantCity")
    country = f.String("applicantCountry")
    # group = f.String("applicantGroup")
    name = f.String("applicantName")
    state = f.String("applicantState")
    zip_code = f.String("applicantZipCode")
    authority_type = f.String("applicantAuthorityType")


class AssigneeSchema(ZipSchema):
    city = f.String("assigneeCity")
    country = f.String("assigneeCountry")
    name = f.String("assigneeName")
    postal_code = f.String("assigneePostalCode")
    state = f.String("assigneeState")
    type_code = f.String("assigneeTypeCode")


class CpcCodeSchema(RegexSchema):
    __regex__ = r"(?P<cpc_class>.{4})(?P<cpc_subclass>[^ ]+) (?P<version>\d{8})"
    cpc_class = f.Str()
    cpc_subclass = f.Str()
    version = f.Date()


class IntlCodeSchema(RegexSchema):
    __regex__ = r"(?P<intl_class>.{4})(?P<intl_subclass>[^ ]+) (?P<version>\d{8})"
    intl_class = f.Str()
    intl_subclass = f.Str()
    version = f.Date()


class ForeignPriorityApplicationSchema(ZipSchema):
    country = f.Str("priorityClaimsCountry")
    app_filing_date = f.Date("priorityClaimsDate")
    app_number = f.Str("priorityClaimsDocNumber")


class PublicSearchDocumentSchema(Schema):
    guid = f.String("guid")
    publication_number = f.String("pubRefDocNumber")
    publication_date = f.Date("datePublished")

    appl_id = f.String("applicationNumber")
    patent_title = f.String("inventionTitle")
    app_filing_date = f.Date("applicationFilingDate.0")
    application_type = f.String("applicationRefFilingType")
    family_identifier_cur = f.Integer("familyIdentifierCur")
    related_apps = RelatedApplicationSchema(data_key=False)
    foreign_priority = ForeignPriorityApplicationSchema(data_key=False)
    type = f.String("type")

    # Parties
    inventors = InventorSchema(data_key=False)
    inventors_short = f.String("inventorsShort")
    applicants = ApplicantSchema(data_key=False)
    assignees = AssigneeSchema(data_key=False)

    group_art_unit = f.String("examinerGroup")
    primary_examiner = f.String("primaryExaminer")
    assistant_examiner = f.List(f.String, "assistantExaminer")
    legal_firm_name = f.List(f.String, "legalFirmName")
    attorney_name = f.List(f.String, "attorneyName")

    # Text Data
    document = DocumentSchema(data_key=False)
    document_structure = DocumentStructureSchema(data_key=False)

    # Image Data
    image_file_name = f.String("imageFileName")
    image_location = f.String("imageLocation")

    # Metadata
    composite_id = f.String("compositeId")
    database_name = f.String("databaseName")
    derwent_week_int = f.Integer("derwentWeekInt")

    # References Cited
    us_references = UsReferenceSchema(data_key=False)
    foreign_references = ForeignReferenceSchema(data_key=False)
    npl_references = f.DelimitedString(NplReferenceSchema, "otherRefPub.0", delimeter="<br />")

    # Classifications
    cpc_inventive = f.List(CpcCodeSchema)
    cpc_additional = f.List(CpcCodeSchema)

    intl_class_issued = f.DelimitedString(f.String, "ipcCodeFlattened", delimeter=";")
    intl_class_current_primary = f.List(IntlCodeSchema, "curIntlPatentClassificationPrimary")
    intl_class_currrent_secondary = f.List(IntlCodeSchema, "curIntlPatentClassificationSecondary")

    us_class_current = f.DelimitedString(f.Str(), "uspcFullClassificationFlattened", delimeter=";")
    us_class_issued = f.List(f.Str, "issuedUsClassificationFull")

    field_of_search_us = f.List(f.Str(), "fieldOfSearchClassSubclassHighlights")
    field_of_search_cpc = f.List(f.Str(), "fieldOfSearchCpcClassification")



from .model import DocumentList
from .model import GlobalDossier
from .session import session


class GlobalDossierApi:
    async def get_file(self, doc_number, type_code="application", office_code="US"):
        url = f"https://gd-api2.uspto.gov/patent-family/svc/family/{type_code}/{office_code}/{doc_number}"
        pre_flight = await session.options(url)
        response = await session.get(url)
        response.raise_for_status()
        return GlobalDossier.model_validate_json(response.content)

    async def get_doc_list(self, country, doc_number, kind_code):
        url = f"https://gd-api2.uspto.gov/doc-list/svc/doclist/{country}/{doc_number}/{kind_code}"
        response = await session.get(url)
        response.raise_for_status()
        return DocumentList.model_validate_json(response.content)

    async def get_document(self, country, doc_number, document_id, out_path):
        url = f"https://gd-api2.uspto.gov/doc-content/svc/doccontent/{country}/{doc_number}/{document_id}/1/PDF"
        return await session.download(url, path=out_path)



import datetime
from pathlib import Path
from typing import Iterable
from typing import List
from typing import Optional
from typing import TYPE_CHECKING

from dateutil.relativedelta import relativedelta
from pydantic import AliasPath
from pydantic import BeforeValidator
from pydantic import ConfigDict
from pydantic import Field
from pydantic import model_validator
from pydantic.alias_generators import to_camel
from typing_extensions import Annotated
from typing_extensions import Self

from .session import session
from patent_client.util.asyncio_util import run_sync
from patent_client.util.pydantic_util import BaseModel
from patent_client.util.pydantic_util import Date
from patent_client.util.pydantic_util import DateTime
from patent_client.util.related import get_model

if TYPE_CHECKING:
    from patent_client import PtabProceeding, Patent, PublishedApplication
    from patent_client.epo.ops.published.model import InpadocBiblio


MDYDate = Annotated[datetime.date, BeforeValidator(lambda x: datetime.datetime.strptime(x, "%m-%d-%Y").date())]
YNBool = Annotated[bool, BeforeValidator(lambda x: x == "Y")]
OptionalInt = Annotated[Optional[int], BeforeValidator(lambda x: int(x) if x else None)]
RelationshipStr = Annotated[str, BeforeValidator(lambda x: x.replace("This application ", "").strip())]


class PEDSBaseModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        str_strip_whitespace=True,
    )


class Transactions(PEDSBaseModel):
    record_date: Date
    code: str
    description: str


class PtaOrPteTransactionHistory(PEDSBaseModel):
    number: float
    pta_or_pte_date: MDYDate
    contents_description: str
    pto_days: OptionalInt
    appl_days: OptionalInt
    start: float


class PtaOrPteSummary(PEDSBaseModel):
    a_delay: int
    b_delay: int
    c_delay: int
    pto_adjustments: int
    total_days: int = Field(alias="totalPtoDays")
    kind: str = Field(alias="ptaPteInd")
    pto_delay: int
    applicant_delay: int = Field(alias="applDelay")
    overlap_delay: int


class ParentApplication(PEDSBaseModel):
    parent_appl_id: str = Field(alias="claimApplicationNumberText")
    child_appl_id: str = Field(alias="applicationNumberText")
    parent_app_filing_date: MDYDate = Field(alias="filingDate")
    parent_patent_number: str = Field(alias="patentNumberText")
    parent_status: str = Field(alias="applicationStatus")
    relationship: RelationshipStr = Field(alias="applicationStatusDescription")


class ChildApplication(PEDSBaseModel):
    child_appl_id: str = Field(alias="claimApplicationNumberText")
    parent_appl_id: str = Field(alias="applicationNumberText")
    child_app_filing_date: MDYDate = Field(alias="filingDate")
    child_patent_number: str = Field(alias="patentNumberText")
    child_status: str = Field(alias="applicationStatus")
    relationship: RelationshipStr = Field(alias="applicationStatusDescription")


class ForeignPriority(PEDSBaseModel):
    priority_claim: str
    country_name: str
    filing_date: MDYDate


class AttorneyOrAgent(PEDSBaseModel):
    application_id: Optional[str] = None
    registration_no: Optional[str] = None
    full_name: str
    phone_num: str
    reg_status: str


class Inventor(PEDSBaseModel):
    name: str
    address: str
    rank_no: int


class Applicant(Inventor):
    cust_no: str = None


class USApplication(PEDSBaseModel):
    appl_id: str
    app_filing_date: Date
    app_exam_name: Optional[str] = None
    public_ind: YNBool
    inventor_name: Optional[str] = None
    app_early_pub_number: Optional[str] = None
    app_early_pub_date: Optional[Date] = None
    patent_number: Optional[str] = None
    patent_issue_date: Optional[Date] = None
    app_location: Optional[str] = None
    app_grp_art_number: Optional[str] = None
    last_modified: DateTime = Field(alias="LAST_MOD_TS")
    last_insert_time: DateTime = Field(alias="LAST_INSERT_TIME")
    patent_title: Optional[str] = None

    app_attr_dock_number: Optional[str] = None
    app_status: Optional[str] = None
    app_status_date: Optional[Date] = None
    app_type: Optional[str] = None
    app_cls_sub_cls: Optional[str] = None
    corr_name: Optional[str] = None
    corr_address: Optional[str] = None
    corr_cust_no: Optional[str] = Field(alias="corrAddrCustNo", default=None)
    transactions: List[Transactions]
    attorneys: List[AttorneyOrAgent] = Field(alias="attrnyAddr", default_factory=list)
    inventors: List[Inventor]
    applicants: List[Applicant] = Field(default_factory=list)
    pta_pte_summary: Optional[PtaOrPteSummary] = None
    pta_pte_tran_history: List[PtaOrPteTransactionHistory] = Field(default_factory=list)
    parent_continuity: List[ParentApplication] = Field(default_factory=list)
    child_continuity: List[ChildApplication] = Field(default_factory=list)
    foreign_priority: List[ForeignPriority] = Field(default_factory=list)

    def __repr__(self):
        return f"USApplication(appl_id='{self.appl_id}', patent_title='{self.patent_title}', app_status='{self.app_status}')"

    @property
    def patent_center_link(self) -> str:
        return f"https://patentcenter.uspto.gov/applications/{self.appl_id}"

    @property
    def google_patents_link(self) -> Optional[str]:
        if self.patent_number is not None:
            return f"https://patents.google.com/patent/US{self.patent_number}/en"
        elif self.app_early_pub_number is not None:
            return f"https://patents.google.com/patent/{self.app_early_pub_number}/en"
        else:
            return None

    @property
    def continuity(self) -> List[Self]:
        """Returns a complete set of parents, self, and children"""
        return list(
            [
                *self.child_continuity.values_list("child", flat=True),
                self,
                *self.parent_continuity.values_list("parent", flat=True),
            ]
        )

    @property
    def kind(self) -> str:
        """Differentiates provisional, PCT, and nonprovisional applications"""
        if "PCT" in self.appl_id:
            return "PCT"
        if self.appl_id[0] == "6":
            return "Provisional"
        return "Nonprovisional"

    @property
    def publication_number(self):
        return self.app_early_pub_number[2:-2]

    @property
    def priority_date(self) -> Optional[datetime.date]:
        """Attempts to return the priority date of the application, calculated as
        the earliest application filing date among the application's parents, or
        its own filing date if it has no parents. Does not include foreign priority
        """
        if not self.parent_continuity:
            return self.app_filing_date
        else:
            return sorted(p.parent_app_filing_date for p in self.parent_continuity)[0]

    @property
    def expiration(self) -> Optional["Expiration"]:
        """Calculates expiration data from which the expiration date can be calculated. See
        help information for the resulting Expiration model.
        """
        if "PCT" in self.appl_id:
            raise NotImplementedError("Expiration date not supported for PCT Applications")
        if not self.patent_number:
            return None
        expiration_data = dict()
        term_parents = [
            p
            for p in self.parent_continuity
            if p.relationship not in ["Claims Priority from Provisional Application", "is a Reissue of"]
        ]
        if term_parents:
            term_parent = sorted(term_parents, key=lambda x: x.parent_app_filing_date)[0]
            relationship = term_parent.relationship
            parent_filing_date = term_parent.parent_app_filing_date
            parent_appl_id = term_parent.parent_appl_id
        else:
            relationship = "self"
            parent_appl_id = self.appl_id
            parent_filing_date = self.app_filing_date

        expiration_data["parent_appl_id"] = parent_appl_id
        expiration_data["parent_app_filing_date"] = parent_filing_date
        expiration_data["parent_relationship"] = relationship
        expiration_data["initial_term"] = parent_filing_date + relativedelta(years=20)  # type: ignore
        expiration_data["pta_or_pte"] = self.pta_pte_summary.total_days or 0  # type: ignore
        expiration_data["extended_term"] = expiration_data["initial_term"] + relativedelta(
            days=expiration_data["pta_or_pte"]
        )  # type: ignore

        transactions = self.transactions
        try:
            disclaimer = next(t for t in transactions if t.code == "DIST")
            expiration_data["terminal_disclaimer_filed"] = True
        except StopIteration:
            expiration_data["terminal_disclaimer_filed"] = False

        return Expiration(**expiration_data)  # type: ignore

    # Related objects
    @property
    def documents(self) -> "Iterable[Document]":
        """File History Documents from PEDS CMS"""
        return get_model("patent_client.uspto.peds.model.Document").objects.filter(appl_id=self.appl_id)

    @property
    def related_assignments(
        self,
    ) -> "Iterable[Assignment]":
        """Related Assignments from the Assignments API"""
        return get_model("patent_client.uspto.assignment.model.Assignment").objects.filter(appl_id=self.appl_id)

    @property
    def ptab_proceedings(
        self,
    ) -> "Iterable[PtabProceeding]":
        """Related PtabProceedings for this application"""
        return get_model("patent_client.uspto.ptab.model.PtabProceeding").objects.filter(appl_id=self.appl_id)

    @property
    def patent(self) -> "Optional[Patent]":
        """Fulltext version of the patent - If Available"""
        return get_model("patent_client.uspto.fulltext.patent.model.Patent").objects.get(
            publication_number=self.patent_number
        )

    @property
    def publication(
        self,
    ) -> "Optional[PublishedApplication]":
        """Fulltext version of the Publication - If Available"""
        return get_model("patent_client.uspto.public_search.model.PublishedApplication").objects.get(
            publication_number=self.publication_number,
        )

    @property
    def inpadoc_patent(
        self,
    ) -> "Optional[InpadocBiblio]":
        """Fulltext version of the patent - If Available"""
        return get_model("patent_client.epo.ops.published.model.InpadocBiblio").objects.get(
            publication_number=f"US{self.patent_number}",
        )

    @property
    def inpadoc_publication(
        self,
    ) -> "Optional[InpadocBiblio]":
        """Fulltext version of the patent - If Available"""
        return get_model("patent_client.epo.ops.published.model.InpadocBiblio").objects.get(
            publication_number=self.app_early_pub_number,
        )

    @model_validator(mode="before")
    @classmethod
    def collect_related_fields(cls, values) -> dict:
        # Collect Correspondent Data
        correspondent_name_fields = ["corrAddrNameLineOne", "corrAddrNameLineTwo", "corrAddrNameLineThree"]
        values["corrName"] = " ".join([values[field] for field in correspondent_name_fields if field in values])
        correspondent_adddress_lines = "\n".join(
            values[f] for f in ("corrAddrStreetLineOne", "corrAddrStreetLineTwo") if f in values
        )
        correspondent_address_last_line = f"{values.get('corrAddrCity', '')}, {values.get('corrAddrGeoRegionCode', '')} {values.get('corrAddrPostalCode', '')}"
        if "corrAddrCountryName" in values:
            correspondent_address_last_line += f" ({values['corrAddrCountryName']})"
        values["corrAddress"] = "\n".join([correspondent_adddress_lines, correspondent_address_last_line])
        # Collect PTA/PTE Data
        pta_pte_name_fields = (
            "totalPtoDays",
            "ptaPteInd",
            "applDelay",
            "cDelay",
            "ptoAdjustments",
            "overlapDelay",
            "aDelay",
            "bDelay",
            "ptoDelay",
        )
        pta_pte_summary = {k: values[k] for k in pta_pte_name_fields if k in values}
        if pta_pte_summary:
            values["ptaPteSummary"] = pta_pte_summary
        # Collect Inventor Names and Addresses
        for inventor in values["inventors"]:
            inventor["name"] = f"{inventor['nameLineOne']}; {inventor['nameLineTwo']} {inventor['suffix']}".strip()
            inventor["address"] = (
                "\n".join([inventor["streetOne"], inventor["streetTwo"]])
                + "\n"
                + f"{inventor['city']}{inventor['geoCode']} {inventor['country']}"
            )
        for applicant in values.get("applicants", list()):
            applicant["name"] = f"{applicant['nameLineOne']}; {applicant['nameLineTwo']} {applicant['suffix']}".strip()
            applicant["address"] = (
                "\n".join([applicant["streetOne"], applicant["streetTwo"]])
                + "\n"
                + f"{applicant['city']}{applicant['geoCode']} {applicant['country']}"
            )

        return values


class Expiration(BaseModel):
    """Model for patent application expiration data.
    NOTE: THIS IS NOT LEGAL ADVICE. See MPEP Sec. 2701 for a detailed explanation of calculating patent term

    This model provides a first-cut estimate of patent term by calulating four things:

    1. The earliest-filed non-provisional patent application (EFNP) from which term should be calculated (parent).
    2. 20 years from the filing date of the EFNP (initial term)
    3. Extensions from Patent Term Extentions (PTE) or Patent Term Adjustments (PTA) (extended_term)
    4. The presence or absence of a terminal disclaimer in the transaction history

    """

    parent_appl_id: str
    """Application number for the earliest-filed nonprovisional application related to this application, or self"""
    parent_app_filing_date: Date
    """Filing date of the earliest-filed nonprovisional application related to this application, or self"""
    parent_relationship: str
    """Relationship of the earliest-filed nonprovisional application. Can be self"""
    initial_term: Date
    """Patent term calculated as 20 years from earliest-field non-provisional (no adjustments)"""
    pta_or_pte: float
    """Days of extended patent term from a Patent Term Extension (PTE) or Patent Term Adjustment (PTA)"""
    extended_term: Date
    """Patent term as extended by any applicable Patent Term Extension or Patent Term Adjustment"""
    terminal_disclaimer_filed: bool
    """Presence or absence of a terminal disclaimer in the transaction history of the application"""


class Document(PEDSBaseModel):
    application_number_text: str
    mail_room_date: Date
    document_code: str
    document_description: str
    document_category: str
    access_level_category: str
    document_identifier: str
    page_count: int
    pdf_url: Optional[str] = None

    def download(self, path: Optional[str | Path]) -> Path:
        return run_sync(self.adownload(path=path))

    async def adownload(self, path: Optional[str | Path]) -> Path:
        if self.pdf_url is None:
            raise ValueError("No PDF URL available")
        full_url = f"https://ped.uspto.gov/api/queries/cms/{self.pdf_url}"
        out_path = await session.adownload(full_url, path=path)
        return out_path

    @property
    def application(self) -> USApplication:
        return run_sync(self.aapplication)

    @property
    async def aapplication(self) -> USApplication:
        return await get_model("patent_client.uspto.peds.model.USApplication").objects.aget(
            appl_id=self.application_number_text
        )


class PedsPage(PEDSBaseModel):
    num_found: int = Field(validation_alias=AliasPath("queryResults", "searchResponse", "response", "numFound"))
    applications: List[USApplication] = Field(
        validation_alias=AliasPath("queryResults", "searchResponse", "response", "docs")
    )



import json
from pathlib import Path

import lxml.etree as ET

from .schema import FamilySchema
from patent_client.util.test import compare_dicts

fixtures = Path(__file__).parent / "fixtures"


def test_example():
    tree = ET.parse(fixtures / "family_input.xml")
    result = FamilySchema().load(tree)
    expected_file = fixtures / "family_convert.json"
    # expected_file.write_text(result.to_json(indent=2))
    expected = json.loads(expected_file.read_text())
    compare_dicts(json.loads(result.to_json()), expected)



import json
from pathlib import Path

import lxml.etree as ET

from .schema import LegalSchema
from patent_client.util.test import compare_dicts

fixture_dir = Path(__file__).parent / "fixtures"


def test_example_1():
    tree = ET.parse(fixture_dir / "example_1.xml")
    result = LegalSchema().load(tree)
    expected_file = fixture_dir / "example_1_convert.json"
    # expected_file.write_text(result.to_json(indent=2))
    expected = json.loads(expected_file.read_text())
    compare_dicts(json.loads(result.to_json()), expected)


def test_example_2():
    tree = ET.parse(fixture_dir / "example_2.xml")
    result = LegalSchema().load(tree)
    expected_file = fixture_dir / "example_2_convert.json"
    # expected_file.write_text(result.to_json(indent=2))
    expected = json.loads(expected_file.read_text())
    compare_dicts(json.loads(result.to_json()), expected)


def test_example_3():
    tree = ET.parse(fixture_dir / "example_3.xml")
    result = LegalSchema().load(tree)
    expected_file = fixture_dir / "example_3_convert.json"
    # expected_file.write_text(result.to_json(indent=2))
    expected = json.loads(expected_file.read_text())
    compare_dicts(json.loads(result.to_json()), expected)



import datetime
import typing as tp
from pathlib import Path

from pydantic import Field

from patent_client.util.asyncio_util import run_sync
from patent_client.util.pydantic_util import BaseModel


class File(BaseModel):
    link_path: str = Field(alias="fileLinkPath")
    identifier: int = Field(alias="fileIdentifier")
    name: str = Field(alias="fileName")
    size: int = Field(alias="fileSize")
    download_url: str = Field(alias="fileDownloadUrl")
    from_date: datetime.date = Field(alias="fileFromTime")
    to_date: datetime.date = Field(alias="fileToTime")
    type: str = Field(alias="fileType")
    release_date: datetime.date = Field(alias="fileReleaseDate")

    async def adownload(self, path: tp.Optional[str | Path]):
        from .session import session

        return await session.adownload(self.download_url, path=path)

    def download(self, path: tp.Optional[str | Path] = None):
        return run_sync(self.adownload(path))


class Product(BaseModel):
    link_path: str = Field(alias="productLinkPath")
    identifier: int = Field(alias="productIdentifier")
    short_name: str = Field(alias="productShortName")
    description: str = Field(alias="productDesc")
    title: str = Field(alias="productTitle")
    frequency: tp.Optional[str] = Field(alias="productFrequency", default=None)
    level: str = Field(alias="productLevel")
    from_date: datetime.date = Field(alias="productFromDate")
    to_date: datetime.date = Field(alias="productToDate")
    number_of_files: int = Field(alias="numberOfFiles")
    parent_product: tp.Optional[str] = Field(alias="parentProduct", default=None)
    files: tp.Optional[tp.List[File]] = Field(alias="productFiles", default_factory=list)



import calendar
import datetime
import typing as tp

from .api import BulkDataApi
from patent_client.util.asyncio_util import run_async_iterator
from patent_client.util.asyncio_util import run_sync
from patent_client.util.manager import Manager

if tp.TYPE_CHECKING:
    from .model import File, Product


def date_ranges(start_date: datetime.date, end_date: datetime.date):
    # First range: start date to end of month
    end_of_month = datetime.datetime(
        start_date.year, start_date.month, calendar.monthrange(start_date.year, start_date.month)[1]
    )
    yield (start_date, end_of_month.date())

    # Full months between start and end date
    current_month = start_date.replace(day=1) + datetime.timedelta(days=32)
    while current_month.replace(day=1) < end_date.replace(day=1):
        last_day_of_month = current_month.replace(day=calendar.monthrange(current_month.year, current_month.month)[1])
        yield (current_month.replace(day=1), last_day_of_month)
        current_month += datetime.timedelta(days=32)

    # Last range: start of month to end date
    yield (end_date.replace(day=1), end_date)


class ProductManager(Manager):
    def filter_by_latest(self) -> tp.Iterator["Product"]:
        """Returns all products with Latest Files"""
        for item in run_async_iterator(self.afilter_by_latest()):
            yield item

    async def afilter_by_latest(self) -> tp.AsyncIterator["Product"]:
        """Returns all products with Latest Files"""
        result = await BulkDataApi.get_latest()
        for product in result:
            yield product

    def get_by_short_name(self, short_name) -> "Product":
        return run_sync(self.aget_by_short_name(short_name))

    async def aget_by_short_name(self, short_name) -> "Product":
        return await BulkDataApi.get_by_short_name(short_name)

    def filter_by_name(self, short_name) -> tp.Iterator["Product"]:
        for item in run_async_iterator(self.afilter_by_name(short_name)):
            yield item

    async def afilter_by_name(self, short_name) -> tp.AsyncIterator["Product"]:
        result = await BulkDataApi.get_by_name(short_name)
        for product in result:
            yield product


class FileManager(Manager):
    def filter_by_short_name(self, short_name, from_date=None, to_date=None) -> tp.Iterator["File"]:
        for item in run_async_iterator(self.afilter_by_short_name(short_name, from_date, to_date)):
            yield item

    async def afilter_by_short_name(self, short_name, from_date=None, to_date=None) -> tp.AsyncIterator["File"]:
        if from_date is not None:
            from_date = from_date if isinstance(from_date, datetime.date) else datetime.date.fromisoformat(from_date)
        if to_date is not None:
            to_date = to_date if isinstance(to_date, datetime.date) else datetime.date.fromisoformat(to_date)
        if from_date is None or to_date is None:
            product = await BulkDataApi.get_by_short_name(short_name)
            from_date = from_date or product.from_date
            to_date = to_date or product.to_date
        for start_date, end_date in date_ranges(from_date, to_date):
            chunk = await BulkDataApi.get_by_short_name(short_name, from_date=start_date, to_date=end_date)
            for file in chunk.files:
                yield file



import datetime
import typing as tp

from .model import Product
from .session import session


class BulkDataApi:
    @classmethod
    async def get_latest(cls) -> tp.List[Product]:
        """Returns all products with Latest Files"""
        response = await session.get("https://bulkdata.uspto.gov:443/BDSS-API/products/all/latest")
        return [Product(**product) for product in response.json()]

    @classmethod
    async def get_by_name(
        cls,
        product_name: str,
        from_date: tp.Optional[datetime.date] = None,
        to_date: tp.Optional[datetime.date] = None,
        max_files: int = 20,
    ) -> tp.List[Product]:
        """Returns files associated with products (of level PRODUCT) based on their full or partial names."""
        params = dict(hierarchy=False)
        if from_date:
            if isinstance(from_date, str):
                from_date = datetime.date.fromisoformat(from_date)
            params["fromYear"] = from_date.year
            params["fromMonth"] = from_date.month
            params["fromDay"] = from_date.day
        if to_date:
            if isinstance(to_date, str):
                to_date = datetime.date.fromisoformat(to_date)
            params["toYear"] = to_date.year
            params["toMonth"] = to_date.month
            params["toDay"] = to_date.day
        params["maxFiles"] = max_files
        response = await session.get(
            f"https://bulkdata.uspto.gov:443/BDSS-API/products/byname/{product_name}", params=params
        )
        response.raise_for_status()
        return [Product.model_validate(product) for product in response.json()]

    @classmethod
    async def get_popular(cls) -> tp.List[Product]:
        """Returns popular products along with latest files."""
        response = await session.get("https://bulkdata.uspto.gov:443/BDSS-API/products/popular")
        return [Product.model_validate(product) for product in response.json()]

    @classmethod
    async def get_by_short_name(
        cls,
        product_name: str,
        from_date: tp.Optional[datetime.date | str] = None,
        to_date: tp.Optional[datetime.date | str] = None,
        max_files: int = 20,
    ) -> tp.List[Product]:
        """Returns files associated with products (of level PRODUCT) based on their full or partial names."""
        params = dict()
        if from_date:
            if isinstance(from_date, str):
                from_date = datetime.date.fromisoformat(from_date)
            params["fromYear"] = from_date.year
            params["fromMonth"] = from_date.month
            params["fromDay"] = from_date.day
        if to_date:
            if isinstance(to_date, str):
                to_date = datetime.date.fromisoformat(to_date)
            params["toYear"] = to_date.year
            params["toMonth"] = to_date.month
            params["toDay"] = to_date.day
        response = await session.get(f"https://bulkdata.uspto.gov:443/BDSS-API/products/{product_name}", params=params)
        return Product.model_validate(response.json())



from typing import List
from typing import Optional

from pydantic import computed_field
from pydantic import Field

from ..schema.biblio import BiblioResultSchema
from patent_client.epo.ops.number_service.model import DocumentId
from patent_client.epo.ops.util import EpoBaseModel
from patent_client.epo.ops.util import InpadocModel


class Citation(InpadocModel):
    cited_phase: Optional[str] = None
    cited_by: Optional[str] = None
    epodoc: Optional[DocumentId] = None
    docdb: Optional[DocumentId] = None
    original: Optional[DocumentId] = None

    def __repr__(self):
        return f"Citation(doc_number={str(self.docdb)}, cited_by={self.cited_by}, cited_phase={self.cited_phase})"

    @computed_field
    @property
    def docdb_number(self) -> str:
        num = self.docdb
        if num.kind:
            return f"{num.country}{num.number}.{num.kind}"
        return f"{num.country}{num.number}"


class Title(EpoBaseModel):
    lang: str
    text: str

    def __repr__(self) -> str:
        return f"Title(lang={self.lang}, text={self.text})"


def limit_text(string, limit=30):
    if len(string) < limit:
        return string
    else:
        return f"{string[:limit]}..."


class InpadocBiblio(InpadocModel):
    __manager__ = "patent_client.epo.ops.published.manager.BiblioManager"
    country: Optional[str] = None
    doc_number: Optional[str] = None
    kind: Optional[str] = None
    family_id: Optional[str] = None
    publication_number: Optional[str] = None
    application_number: Optional[str] = None
    publication_reference_docdb: Optional[DocumentId] = None
    publication_reference_epodoc: Optional[DocumentId] = None
    publication_reference_original: Optional[DocumentId] = None
    application_reference_docdb: Optional[DocumentId] = None
    application_reference_epodoc: Optional[DocumentId] = None
    application_reference_original: Optional[DocumentId] = None
    intl_class: List[str] = Field(default_factory=list)
    cpc_class: List[str] = Field(default_factory=list)
    us_class: List[str] = Field(default_factory=list)
    priority_claims: List[DocumentId] = Field(default_factory=list)
    title: Optional[str] = None
    titles: List[Title] = Field(default_factory=list)
    abstract: Optional[str] = None
    citations: List[Citation] = Field(default_factory=list)
    applicants_epodoc: List[str] = Field(default_factory=list)
    applicants_original: List[str] = Field(default_factory=list)
    inventors_epodoc: List[str] = Field(default_factory=list)
    inventors_original: List[str] = Field(default_factory=list)
    # TODO: NPL citations

    def __repr__(self):
        return f"InpadocBiblio(publication_number={self.publication_number}, title={limit_text(self.title, 30)})"


class BiblioResult(EpoBaseModel):
    __schema__ = BiblioResultSchema()
    documents: list[InpadocBiblio] = Field(default_factory=list)



import dotenv

dotenv.load_dotenv()
from pathlib import Path
from typing import Optional
from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="patent_client_")
    base_dir: Path = Field(default=Path("~/.patent_client").expanduser())
    log_file: str = Field(default="patent_client.log")
    log_level: str = Field(default="INFO")
    epo_api_key: Optional[str] = Field(default=None)
    epo_api_secret: Optional[str] = Field(default=None)
    itc_username: Optional[str] = Field(default=None)
    itc_password: Optional[str] = Field(default=None)



import json
from pathlib import Path

import lxml.etree as ET
import pytest

from .schema import BiblioResultSchema
from .schema import ClaimsSchema
from .schema import DescriptionSchema
from .schema import ImagesSchema
from .schema import SearchSchema
from patent_client.util.test import compare_dicts

test_dir = Path(__file__).parent / "test"
expected_dir = Path(__file__).parent / "test" / "expected"


def test_biblio():
    tree = ET.parse(test_dir / "biblio_example.xml")
    result = BiblioResultSchema().load(tree)
    assert len(result.documents) == 1
    d = result.documents[0]
    assert d.title == "Bi-directional steerable catheter control handle"
    assert d.country == "US"
    assert d.doc_number == "2006142694"
    assert d.family_id == "35840242"
    for f in (
        "publication_reference_docdb",
        "publication_reference_epodoc",
        "application_reference_docdb",
        "application_reference_epodoc",
        "application_reference_original",
    ):
        assert hasattr(d, f)
    assert len(d.citations) == 99


@pytest.mark.parametrize("filename", ["claims_example.xml", "claims_example_2.xml"])
def test_claims(filename):
    input_file = test_dir / filename
    tree = ET.parse(input_file)
    result = ClaimsSchema().load(tree)
    example_file = input_file.with_suffix(".json")
    example_file.write_text(result.to_json(indent=2))
    expected = json.loads(example_file.read_text())
    compare_dicts(json.loads(result.to_json()), expected)


def test_description():
    tree = ET.parse(test_dir / "description_example.xml")
    result = DescriptionSchema().load(tree)
    assert result.document_id.country == "EP"
    assert result.document_id.doc_number == "1000000"
    assert result.document_id.kind == "A1"
    assert len(result.description) == 10544


def test_search():
    tree = ET.parse(test_dir / "search_example.xml")
    result = SearchSchema().load(tree)
    assert result.query == 'applicant = "nine energy"'
    assert result.num_results == 15
    assert result.begin == 1
    assert result.end == 25
    assert len(result.results) == 15
    r = result.results[3]
    assert r.family_id == "70280448"
    assert r.id_type == "docdb"
    assert r.country == "US"


def test_images():
    tree = ET.parse(test_dir / "image_example.xml")
    result = ImagesSchema().load(tree)
    expected_file = expected_dir / "image_example.json"
    # expected_file.write_text(result.to_json())
    expected = json.loads(expected_file.read_text())
    compare_dicts(json.loads(result.to_json()), expected)


def test_full_cycle():
    tree = ET.parse(test_dir / "full_cycle_example.xml")
    result = BiblioResultSchema().load(tree)
    assert len(result.documents) == 2
    d = result.documents[0]
    assert d.title == "Apparatus for manufacturing green bricks for the brick manufacturing industry"


def test_biblio_2():
    filename = "biblio_example_4"
    input = test_dir / f"{filename}.xml"
    tree = ET.parse(input)
    result = BiblioResultSchema().load(tree)
    expected_file = expected_dir / f"{filename}.json"
    # expected_file.write_text(result.to_json(indent=2))
    compare_dicts(json.loads(result.to_json()), json.loads(expected_file.read_text()))



from yankee.json import fields as f
from yankee.json import Schema


class DocumentStructureSchema(Schema):
    number_of_claims = f.Integer("numberOfClaims")
    number_of_drawing_sheets = f.Integer("numberOfDrawingSheets")
    number_of_figures = f.Integer("numberOfFigures")
    page_count = f.Integer("pageCount")
    front_page_end = f.Integer("frontPageEnd")
    front_page_start = f.Integer("frontPageStart")
    bib_start = f.Integer("bibStart")
    bib_end = f.Integer("bibEnd")
    abstract_start = f.Integer("abstractStart")
    abstract_end = f.Integer("abstractEnd")
    drawings_start = f.Integer("drawingsStart")
    drawings_end = f.Integer("drawingsEnd")
    description_start = f.Integer("descriptionStart")
    description_end = f.Integer("descriptionEnd")
    specification_start = f.Integer("specificationStart")
    specification_end = f.Integer("specificationEnd")
    claims_end = f.Integer("claimsEnd")
    claims_start = f.Integer("claimsStart")
    amend_start = f.Integer("amendStart")
    amend_end = f.Integer("amendEnd")
    cert_correction_end = f.Integer("certCorrectionEnd")
    cert_correction_start = f.Integer("certCorrectionStart")
    cert_reexamination_end = f.Integer("certReexaminationEnd")
    cert_reexamination_start = f.Integer("certReexaminationStart")
    ptab_start = f.Integer("ptabStart")
    ptab_end = f.Integer("ptabEnd")
    search_report_start = f.Integer("searchReportStart")
    search_report_end = f.Integer("searchReportEnd")
    supplemental_start = f.Integer("supplementalStart")
    supplemental_end = f.Integer("supplementalEnd")



from typing import List
from typing import Optional
from typing import TYPE_CHECKING

from pydantic import computed_field
from pydantic import model_validator
from yankee.xml.schema import Schema as XmlSchema

from patent_client.util.pydantic_util import BaseModel
from patent_client.util.related import get_model

if TYPE_CHECKING:
    from patent_client.epo.ops.published.model import InpadocBiblio
    from patent_client.epo.ops.published.model import Images
    from patent_client.epo.ops.published.model import Claims
    from patent_client.epo.ops.legal.model import LegalEvent
    from patent_client.epo.ops.family.model import FamilyMember


class Schema(XmlSchema):
    class Meta:
        namespaces = {
            "ops": "http://ops.epo.org",
            "epo": "http://www.epo.org/exchange",
            "ft": "http://www.epo.org/fulltext",
        }


class EpoBaseModel(BaseModel):
    __schema__: Optional[XmlSchema] = None

    @model_validator(mode="before")
    @classmethod
    def xml_convert(cls, values):
        if isinstance(values, (str, bytes)):
            return cls.__schema__.load(values).to_dict()
        return values


class InpadocModel(EpoBaseModel):
    @computed_field
    @property
    def docdb_number(self) -> str:
        num = self.publication_reference_docdb
        if num.kind:
            return f"{num.country}{num.number}.{num.kind}"
        return f"{num.country}{num.number}"

    @property
    def biblio(self) -> "InpadocBiblio":
        return get_model("patent_client.epo.ops.published.model.InpadocBiblio").objects.get(self.docdb_number)

    @property
    def images(self) -> "Images":
        return get_model("patent_client.epo.ops.published.model.Images").objects.get(self.docdb_number)

    @property
    def description(self) -> Optional[str]:
        return get_model("patent_client.epo.ops.published.model.Description").objects.get(self.docdb_number).description

    @property
    def claims(self) -> "Claims":
        return get_model("patent_client.epo.ops.published.model.Claims").objects.get(self.docdb_number)

    @property
    def legal(self) -> List["LegalEvent"]:
        return get_model("patent_client.epo.ops.legal.model.Legal").objects.get(self.docdb_number).events

    @property
    def family(
        self,
    ) -> List["FamilyMember"]:
        return get_model("patent_client.epo.ops.family.model.Family").objects.get(self.docdb_number).family_members



from typing import AsyncIterator
from typing import Generic
from typing import TypeVar

from .api import PublicSearchApi
from .model import Patent
from .model import PatentBiblio
from .model import PublicSearchBiblio
from .model import PublicSearchDocument
from .model import PublishedApplication
from .model import PublishedApplicationBiblio
from .query import QueryBuilder
from patent_client.util.manager import Manager
from patent_client.util.request_util import get_start_and_row_count


class CapacityException(Exception):
    pass


class FinishedException(Exception):
    pass


T = TypeVar("T")

public_search_api = PublicSearchApi()


class GenericPublicSearchManager(Manager, Generic[T]):
    page_size = 500
    default_filter = "patent_number"
    query_builder = QueryBuilder()

    @property
    def _query(self):
        return self.query_builder.build_query(self.config)

    @property
    def _order_by(self):
        return self.query_builder.build_order_by(self.config)

    @property
    def query_fields(self):
        return self.query_builder.search_keywords

    @property
    def order_by_fields(self):
        return self.query_builder.order_by_keywords

    async def alen(self):
        query = self._query
        order_by = self._order_by
        sources = self.config.options.get("sources", ["US-PGPUB", "USPAT", "USOCR"])
        page = await public_search_api.run_query(
            query=query, start=0, limit=self.page_size, sort=order_by, sources=sources
        )
        max_len = page.num_found - self.config.offset
        return min(self.config.limit, max_len) if self.config.limit else max_len


class GenericPublicSearchBiblioManager(GenericPublicSearchManager, Generic[T]):
    async def _aget_results(self) -> AsyncIterator[T]:
        query = self._query
        order_by = self._order_by
        sources = self.config.options.get("sources", ["US-PGPUB", "USPAT", "USOCR"])
        for start, rows in get_start_and_row_count(self.config.limit, self.config.offset, self.page_size):
            page = await public_search_api.run_query(
                query=query,
                start=start,
                limit=rows,
                sort=order_by,
                sources=sources,
            )
            for obj in page.docs:
                yield obj
            if len(page.docs) < rows:
                break


class GenericPublicSearchDocumentManager(GenericPublicSearchBiblioManager, Generic[T]):
    async def _aget_results(self) -> AsyncIterator["PublicSearchDocument"]:
        result_count = super().__len__()
        if result_count > 20:
            raise CapacityException(
                f"Query would result in more than 20 results! ({result_count} > 20).\nPlease use the associated Biblio method to reduce load on the API (PublicSearch / PatentBiblio / PublishedApplicationBiblio"
            )
        async for obj in super()._aget_results():
            doc = await public_search_api.get_document(obj)
            yield doc


class PublicSearchBiblioManager(GenericPublicSearchManager[PublicSearchBiblio]):
    pass


class PublicSearchDocumentManager(GenericPublicSearchDocumentManager[PublicSearchDocument]):
    pass


class PatentBiblioManager(GenericPublicSearchBiblioManager[PatentBiblio]):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.config.options["sources"] = [
            "USPAT",
        ]


class PatentManager(GenericPublicSearchDocumentManager[Patent]):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.config.options["sources"] = [
            "USPAT",
        ]


class PublishedApplicationBiblioManager(GenericPublicSearchBiblioManager[PublishedApplicationBiblio]):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.config.options["sources"] = [
            "US-PGPUB",
        ]


class PublishedApplicationManager(GenericPublicSearchDocumentManager[PublishedApplication]):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.config.options["sources"] = [
            "US-PGPUB",
        ]



from pathlib import Path

import pytest

from .api import LegalAsyncApi

fixtures = Path(__file__).parent / "fixtures"


@pytest.mark.asyncio
async def test_async_example():
    result = await LegalAsyncApi.get_legal("EP1000000A1")
    assert str(result.publication_reference) == "EP1000000A1"
    assert len(result.events) > 50



import re
import warnings
from contextlib import contextmanager
from pathlib import Path
from typing import Optional
from typing import Union

import hishel
import httpx

from patent_client import CACHE_DIR
from patent_client.util.asyncio_util import run_sync
from patent_client.version import __version__

try:
    from IPython import get_ipython

    if "IPKernelApp" in get_ipython().config:
        from tqdm.notebook import tqdm
    else:
        from tqdm import tqdm
except (ImportError, AttributeError):
    from tqdm import tqdm

filename_re = re.compile(r'filename="([^"]+)"')


class PatentClientAsyncFileCache(hishel.AsyncFileStorage):
    _default_base_path = CACHE_DIR

    def __init__(self, **kwargs):
        kwargs["base_path"] = kwargs.get("base_path", self._default_base_path)
        super(PatentClientAsyncFileCache, self).__init__(**kwargs)

    def delete(self, cache_key: str):
        response_path = self._base_path / cache_key
        if response_path.exists():
            response_path.unlink()


class PatentClientController(hishel.Controller):
    def construct_response_from_cache(
        self, request: httpx.Request, response: httpx.Response, original_request: httpx.Request
    ) -> Union[httpx.Response, httpx.Request, None]:
        try:
            super(PatentClientController, self).construct_response_from_cache(request, response, original_request)
        except RuntimeError:
            self._make_request_conditional(request=request, response=response)
            return request


class PatentClientSession(hishel.AsyncCacheClient):
    _default_transport = httpx.AsyncHTTPTransport(
        retries=3,
        verify=False,
        http2=True,
    )
    _default_timeout = 60 * 5
    _default_user_agent = f"Mozilla/5.0 Python Patent Clientbot/{__version__} (parkerhancock@users.noreply.github.com)"
    _default_storage = PatentClientAsyncFileCache()
    _default_controller = PatentClientController(
        allow_heuristics=True,
    )

    def __init__(self, **kwargs):
        kwargs["transport"] = kwargs.get("transport", self._default_transport)
        kwargs["storage"] = kwargs.get("storage", self._default_storage)
        kwargs["controller"] = kwargs.get("controller", self._default_controller)
        headers = kwargs.get("headers", dict())
        headers["User-Agent"] = headers.get("User-Agent", self._default_user_agent)
        kwargs["headers"] = headers
        kwargs["follow_redirects"] = kwargs.get("follow_redirects", True)
        kwargs["timeout"] = kwargs.get("timeout", self._default_timeout)
        super(PatentClientSession, self).__init__(**kwargs)

    def delete(self, response: httpx.Response):
        if "cache_metadata" in response.extensions:
            self._storage.delete(response.extensions["cache_metadata"]["cache_key"])

    def download(self, url, method: str = "GET", path: Optional[str | Path] = None, **kwargs):
        return run_sync(self.adownload(url, method, path, **kwargs))

    async def adownload(self, url, method: str = "GET", path: Optional[str | Path] = None, **kwargs):
        if isinstance(path, str):
            path = Path(path)
        elif path is None:
            path = Path.cwd()
        if not path.is_dir() and path.exists():
            warnings.warn(
                "File already exists at provided output! Not re-downloading. Please move the file or provide an alternative path to download"
            )
            return path
        async with self.stream(method, url, **kwargs) as response:
            response.raise_for_status()
            total = int(response.headers["Content-Length"])
            with tqdm(total=total, unit_scale=True, unit_divisor=1024, unit="B") as progress:
                num_bytes_downloaded = response.num_bytes_downloaded
                if path.is_dir() or None:
                    try:
                        filename = filename_re.search(response.headers["Content-Disposition"]).group(1)
                    except (AttributeError, KeyError):
                        filename = url.split("/")[-1]
                    path = path / filename if path else Path.cwd() / filename
                with path.open("wb") as f:
                    async for chunk in response.aiter_bytes():
                        f.write(chunk)
                        progress.update(response.num_bytes_downloaded - num_bytes_downloaded)
                        num_bytes_downloaded = response.num_bytes_downloaded
        return path

    @contextmanager
    def cache_disabled(session):
        if hasattr(session._transport, "_transport"):
            cached_transport = session._transport._transport
            session._transport = session._transport._transport
            yield
            session._transport = cached_transport
        else:
            yield



import pytest

from .api import BulkDataApi
from .model import File
from .model import Product


@pytest.mark.asyncio
async def test_can_get_latest():
    latest = await BulkDataApi.get_latest()
    first = latest[0]
    assert isinstance(first, Product)
    assert isinstance(first.files[0], File)


@pytest.mark.asyncio
async def test_can_get_by_short_name():
    first = await BulkDataApi.get_by_short_name("PTGRXML")
    assert isinstance(first, Product)
    assert isinstance(first.files[0], File)


@pytest.mark.asyncio
async def test_can_get_by_short_name_with_date_range():
    first = await BulkDataApi.get_by_short_name("PTGRXML", from_date="2023-06-01", to_date="2023-08-01")
    assert isinstance(first, Product)
    assert isinstance(first.files[0], File)
    assert len(first.files) == 10


@pytest.mark.asyncio
@pytest.mark.no_vcr
@pytest.mark.skip(reason="This test is too slow")
async def test_can_download_file(tmpdir):
    product = await BulkDataApi.get_by_short_name("PTGRXML")
    file = product.files[0]
    await file.adownload(tmpdir)
    assert len(list(tmpdir.listdir())) == 1



from pathlib import Path
from typing import Optional
from typing import TYPE_CHECKING

from yankee.base.schema import ListCollection

from .convert import convert_xml_to_json
from .model import AssignmentPage
from .session import session

if TYPE_CHECKING:
    from .model import Assignment

allowed_filters = [
    "PCTNumber",
    "OwnerName",
    "CorrespondentName",
    "PriorOwnerName",
    "ApplicationNumber",
    "PatentNumber",
    "PublicationNumber",
    "IntlRegistrationNumber",
    "ReelFrame",
]

allowed_sorts = ["ExecutionDate+desc", "ExecutionDate+asc"]


def validate_input(input, input_name, allowed_values):
    if input not in allowed_values:
        raise ValueError(f"{input_name} must be one of {allowed_values}")


class AssignmentApi:
    @classmethod
    async def alookup(
        self, query: str, filter: str, rows=8, start=0, sort="ExecutionDate+desc"
    ) -> ListCollection["Assignment"]:
        # Because of their limited utility, we omit fields, highlight, and facet
        url = "https://assignment-api.uspto.gov/patent/lookup"
        validate_input(filter, "filter", allowed_filters)
        validate_input(sort, "sort", allowed_sorts)
        response = await session.get(
            url,
            params={
                "filter": filter,
                "query": query,
                "rows": rows,
                "start": start,
                "sort": sort,
                "facet": "false",
            },
            headers={"Accept": "application/xml"},
        )
        response.raise_for_status()
        return AssignmentPage.model_validate(convert_xml_to_json(response.content))

    @classmethod
    async def download_pdf(self, reel: str, frame: str, path: Optional[Path] = None) -> Path:
        url = self.get_download_url(reel, frame)

        if path is None:
            path = Path.cwd()
            output_path = output_path = path / f"assignment-pat-{reel}-{frame}.pdf"
        elif path.is_dir():
            output_path = path / f"assignment-pat-{reel}-{frame}.pdf"
        else:
            output_path = path

        with output_path.open("wb") as f:
            async with session.stream("GET", url) as response:
                async for chunk in response.aiter_bytes():
                    f.write(chunk)
        return output_path



from typing import List
from typing import Optional

from pydantic import BeforeValidator
from pydantic import Field
from pydantic import model_validator
from typing_extensions import Annotated

from ..convert.document import PublicSearchDocumentSchema
from ..util import html_to_text
from .shared import ApplicationNumber
from .shared import DocumentStructure
from .shared import OptionalList
from patent_client.util.asyncio_util import run_sync
from patent_client.util.claims.model import Claim
from patent_client.util.pydantic_util import BaseModel
from patent_client.util.pydantic_util import Date
from patent_client.util.related import get_model

HtmlString = Annotated[str, BeforeValidator(html_to_text)]


class Document(BaseModel):
    abstract_html: Optional[str] = None
    abstract: Optional[HtmlString] = Field(alias="abstract_html", default=None)
    government_interest: Optional[str] = None
    background_html: Optional[str] = Field(alias="background_html", default=None)
    background: Optional[HtmlString] = Field(alias="background_html", default=None)
    description_html: Optional[str] = None
    description: Optional[HtmlString] = Field(alias="description_html", default=None)
    brief_html: Optional[str] = None
    brief: Optional[HtmlString] = Field(alias="brief_html", default=None)
    claim_statement: Optional[str] = None
    claims_html: Optional[str] = None
    claims_text: Optional[HtmlString] = Field(alias="claims_html", default=None)
    claims: List[Claim] = Field(default_factory=list)


class UsReference(BaseModel):
    publication_number: Optional[str] = None
    pub_month: Optional[Date] = None
    patentee_name: Optional[str] = None
    cited_by_examiner: Optional[bool] = None


class ForeignReference(BaseModel):
    citation_classification: Optional[str] = None
    citation_cpc: Optional[str] = None
    country_code: Optional[str] = None
    patent_number: Optional[str] = None
    pub_month: Optional[Date] = None
    cited_by_examiner: Optional[bool] = None


class NplReference(BaseModel):
    citation: Optional[str] = None
    cited_by_examiner: Optional[bool] = None


class RelatedApplication(BaseModel):
    child_patent_country: Optional[str] = None
    child_patent_number: Optional[str] = None
    country_code: Optional[str] = None
    filing_date: Optional[Date] = None
    number: Optional[str] = None
    parent_status_code: Optional[str] = None
    patent_issue_date: Optional[Date] = None
    patent_number: Optional[str] = None


class Inventor(BaseModel):
    name: Optional[str] = None
    city: Optional[str] = None
    country: Optional[str] = None
    postal_code: Optional[str] = None
    state: Optional[str] = None


class Applicant(BaseModel):
    city: Optional[str] = None
    country: Optional[str] = None
    # group: Optional[str] = None
    name: Optional[str] = None
    state: Optional[str] = None
    zip_code: Optional[str] = None
    authority_type: Optional[str] = None


class Assignee(BaseModel):
    city: Optional[str] = None
    country: Optional[str] = None
    name: Optional[str] = None
    postal_code: Optional[str] = None
    state: Optional[str] = None
    type_code: Optional[str] = None


class CpcCode(BaseModel):
    cpc_class: Optional[str] = None
    cpc_subclass: Optional[str] = None
    version: Optional[Date] = None


class IntlCode(BaseModel):
    intl_class: Optional[str] = None
    intl_subclass: Optional[str] = None
    version: Optional[Date] = None


class ForeignPriorityApplication(BaseModel):
    country: Optional[str] = None
    app_filing_date: Optional[Date] = None
    app_number: Optional[str] = None


class PublicSearchDocument(BaseModel):
    __schema__ = PublicSearchDocumentSchema()
    guid: Optional[str] = None
    publication_number: Optional[str] = None
    publication_date: Optional[Date] = None

    appl_id: Optional[ApplicationNumber] = None
    patent_title: Optional[str] = None
    app_filing_date: Optional[Date] = None
    application_type: Optional[str] = None
    family_identifier_cur: Optional[int] = None
    related_apps: List[RelatedApplication] = Field(default_factory=list)
    foreign_priority: List[ForeignPriorityApplication] = Field(default_factory=list)
    type: Optional[str] = None

    # Parties
    inventors: List[Inventor] = Field(default_factory=list)
    inventors_short: Optional[str] = None
    applicants: List[Applicant] = Field(default_factory=list)
    assignees: List[Assignee] = Field(default_factory=list)

    group_art_unit: Optional[str] = None
    primary_examiner: Optional[str] = None
    assistant_examiner: Optional[List[str]] = Field(default_factory=list)
    legal_firm_name: Optional[List[str]] = Field(default_factory=list)
    attorney_name: Optional[List[str]] = Field(default_factory=list)

    # Text Data
    document: Optional[Document] = None
    document_structure: Optional[DocumentStructure] = None

    # Image Data
    image_file_name: Optional[str] = None
    image_location: Optional[str] = None

    # Metadata
    composite_id: Optional[str] = None
    database_name: Optional[str] = None
    derwent_week_int: Optional[int] = None

    # References Cited
    us_references: OptionalList[UsReference] = Field(default_factory=list)
    foreign_references: OptionalList[ForeignReference] = Field(default_factory=list)
    npl_references: OptionalList[NplReference] = Field(default_factory=list)

    # Classifications
    cpc_inventive: OptionalList = Field(default_factory=list)
    cpc_additional: OptionalList = Field(default_factory=list)

    intl_class_issued: OptionalList[str] = Field(default_factory=list)
    intl_class_current_primary: OptionalList[IntlCode] = Field(default_factory=list)
    intl_class_currrent_secondary: OptionalList[IntlCode] = Field(default_factory=list)

    us_class_current: OptionalList[str] = Field(default_factory=list)
    us_class_issued: OptionalList[str] = Field(default_factory=list)

    field_of_search_us: OptionalList[str] = Field(default_factory=list)
    field_of_search_cpc: OptionalList[str] = Field(default_factory=list)

    def __repr__(self):
        return f"PublicSearchDocument(publication_number={self.publication_number}, publication_date={self.publication_date}, patent_title={self.patent_title})"

    @model_validator(mode="before")
    @classmethod
    def convert_json(cls, values):
        return PublicSearchBiblioConvert.convert(values)

    @property
    def abstract(self):
        return self.document.abstract

    @property
    def description(self):
        sections = (
            "government_interest",
            "background",
            "brief",
            "description",
        )
        section_text = tuple(getattr(self.document, s) for s in sections)
        return "\n\n".join(s for s in section_text if s)

    @property
    def claims_text(self):
        return self.document.claims

    @property
    def claims(self):
        return self.document.claims

    @property
    def forward_citations(self):
        return get_model("patent_client.uspto.public_search.model.PublicSearchBiblio").objects.filter(
            us_reference=self.publication_number
        )

    def download_images(self, path="."):
        return run_sync(self.adownload_images(path))

    async def adownload_images(self, path="."):
        from ..manager import public_search_api

        return await public_search_api.download_image(self, path)

    @property
    def application(self):
        return get_model("patent_client.uspto.peds.model.USApplication").objects.get(self.appl_id)

    @property
    def global_dossier(self):
        return get_model("patent_client.uspto.global_dossier.model.GlobalDossierApplication").objects.get(self.appl_id)

    @property
    def assignments(self):
        return get_model("patent_client.uspto.assignment.model.Assignment").objects.filter(appl_id=self.appl_id)

    @property
    def inpadoc(self):
        return get_model("patent_client.epo.ops.published.model.Inpadoc").objects.get("US" + self.publication_number)

    @model_validator(mode="before")
    @classmethod
    def convert_json(cls, values):
        return cls.__schema__.deserialize(values).to_dict()



import datetime

from .manager import date_ranges
from .model import File
from .model import Product


class TestProduct:
    def test_can_get_latest(self):
        latest = list(Product.objects.filter_by_latest())
        assert isinstance(latest[0], Product)

    def test_can_get_by_short_name(self):
        first = Product.objects.get_by_short_name("PTGRXML")
        assert isinstance(first, Product)
        assert isinstance(first.files[0], File)

    def test_can_filter_by_name(self):
        first = Product.objects.filter_by_name("Assignment")
        first = list(first)[0]
        assert isinstance(first, Product)
        assert isinstance(first.files[0], File)


class TestFile:
    def test_can_filter_by_short_name(self):
        results = list(File.objects.filter_by_short_name("PTGRXML", from_date="2023-06-15", to_date="2023-08-15"))
        assert len(results) == 10
        assert isinstance(results[0], File)


def test_date_ranges():
    result = list(date_ranges(datetime.date(2020, 1, 15), datetime.date(2021, 2, 15)))
    assert result[0] == (datetime.date(2020, 1, 15), datetime.date(2020, 1, 31))
    assert result[-1] == (datetime.date(2021, 2, 1), datetime.date(2021, 2, 15))
    assert len(result) == 14



from typing import List
from typing import Optional

from pydantic import computed_field
from pydantic import Field

from ...util import InpadocModel
from ..schema.search import SearchSchema
from patent_client.epo.ops.util import EpoBaseModel


class Inpadoc(InpadocModel):
    family_id: Optional[str] = None
    id_type: Optional[str] = None
    country: Optional[str] = None
    doc_number: Optional[str] = None
    kind: Optional[str] = None

    @computed_field
    @property
    def docdb_number(self) -> str:
        return f"{self.country}{self.doc_number}{self.kind}"

    def download(self, path="."):
        return self.images.full_document.download(path)


class Search(EpoBaseModel):
    __schema__ = SearchSchema()
    query: Optional[str] = None
    num_results: Optional[int] = None
    begin: Optional[int] = None
    end: Optional[int] = None
    results: List[Inpadoc] = Field(default_factory=list)



import datetime
from typing import List
from typing import Optional
from typing import TypeVar

from pydantic import BeforeValidator
from pydantic import ConfigDict
from pydantic.alias_generators import to_camel
from typing_extensions import Annotated

from patent_client.util.pydantic_util import BaseModel


class PublicSearchBaseModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        str_strip_whitespace=True,
    )


T = TypeVar("T")


OptionalList = Annotated[List[T], BeforeValidator(lambda x: x if isinstance(x, list) else list())]
DateTimeAsDate = Annotated[
    datetime.date, BeforeValidator(lambda x: x.date() if isinstance(x, datetime.datetime) else x)
]


def format_appl_id(string):
    if string.startswith("D"):
        string = "29" + string[1:]
    return string.replace("/", "")


ApplicationNumber = Annotated[str, BeforeValidator(format_appl_id)]


class DocumentStructure(BaseModel):
    number_of_claims: Optional[int] = None
    number_of_drawing_sheets: Optional[int] = None
    number_of_figures: Optional[int] = None
    page_count: Optional[int] = None
    front_page_end: Optional[int] = None
    front_page_start: Optional[int] = None
    bib_start: Optional[int] = None
    bib_end: Optional[int] = None
    abstract_start: Optional[int] = None
    abstract_end: Optional[int] = None
    drawings_start: Optional[int] = None
    drawings_end: Optional[int] = None
    description_start: Optional[int] = None
    description_end: Optional[int] = None
    specification_start: Optional[int] = None
    specification_end: Optional[int] = None
    claims_end: Optional[int] = None
    claims_start: Optional[int] = None
    amend_start: Optional[int] = None
    amend_end: Optional[int] = None
    cert_correction_end: Optional[int] = None
    cert_correction_start: Optional[int] = None
    cert_reexamination_end: Optional[int] = None
    cert_reexamination_start: Optional[int] = None
    ptab_start: Optional[int] = None
    ptab_end: Optional[int] = None
    search_report_start: Optional[int] = None
    search_report_end: Optional[int] = None
    supplemental_start: Optional[int] = None
    supplemental_end: Optional[int] = None



import datetime as dt
from typing import Optional

from patent_client import SETTINGS
from patent_client.session import PatentClientSession
from patent_client.util.asyncio_util import SyncProxy


NS = {
    "http://ops.epo.org": None,
    "http://www.epo.org/exchange": None,
    "http://www.epo.org/fulltext": None,
    "http://www.epo.org/register": None,
}

import logging

logger = logging.getLogger(__name__)


class OpsAuthenticationError(Exception):
    pass


class OpsForbiddenError(Exception):
    pass


class OpsFairUseError(Exception):
    pass


class OpsAsyncSession(PatentClientSession):
    key: Optional[str] = None
    secret: Optional[str] = None
    expires: dt.datetime
    sync: SyncProxy

    def __init__(self, *args, key: str = "", secret: str = "", **kwargs):
        super(OpsAsyncSession, self).__init__(*args, **kwargs)
        self.key = key
        self.secret = secret
        self.expires = dt.datetime.utcnow()
        self.sync = SyncProxy(self)

    async def request(self, *args, **kwargs):
        response = await super(OpsAsyncSession, self).request(*args, **kwargs)
        if response.status_code in (403, 400):
            self.delete(response)
            auth_response = await self.get_token()
            response = await super(OpsAsyncSession, self).request(*args, **kwargs)
        if response.status_code in (403, 400):
            self.delete(response)
            if "Fair Use policy" in response.text:
                raise OpsFairUseError(f"EPO Fair Use Policy Error!\n{response.text}{response.headers}")
            else:
                raise OpsForbiddenError(f"EPO Request Error!\nStatus Code: {response.status_code}\n{response.text}")
        response.raise_for_status()
        return response

    async def get_token(self):
        auth_url = "https://ops.epo.org/3.2/auth/accesstoken"
        with self.cache_disabled():
            response = await super(OpsAsyncSession, self).request(
                "post",
                auth_url,
                auth=(self.key, self.secret),
                data={"grant_type": "client_credentials"},
            )
            if response.status_code == 401:
                logger.debug(f"EPO Authentication Error!\n{response.text}")
                raise OpsAuthenticationError(
                    "Failed to authenticate with EPO OPS! Please check your credentials. See the setup instructions at https://patent-client.readthedocs.io/en/stable/getting_started.html"
                )
            elif response.status_code == 403:
                logger.debug(f"EPO Forbidden Error\n{response.text}")
                raise OpsForbiddenError("Your EPO Request Failed - Quota Exceeded / Blacklisted / Blocked")
            response.raise_for_status()

            data = response.json()
            self.expires = dt.datetime.fromtimestamp(int(data["issued_at"]) / 1000) + dt.timedelta(
                seconds=int(data["expires_in"])
            )
            self.headers["Authorization"] = f"Bearer {data['access_token']}"
            return response


asession = OpsAsyncSession(key=SETTINGS.epo_api_key, secret=SETTINGS.epo_api_secret)



import lxml.etree as ET


def convert_doc(doc):
    output = dict()
    for e in doc.iterchildren():
        if e.tag in ("str", "date", "int"):
            output[e.attrib["name"]] = None if e.text == "NULL" else e.text
        elif e.tag == "arr":
            output[e.attrib["name"]] = [None if c.text == "NULL" else c.text for c in e.iterchildren()]
    # Collect assignors into a list of dicts
    assignor_fields = ["patAssignorName", "patAssignorExDate", "patAssignorDateAck"]
    output["assignors"] = zip_lists(output, assignor_fields, "assignor")
    # Collect assignees into a list of dicts
    assignee_fields = [
        "patAssigneeName",
        "patAssigneeAddress1",
        "patAssigneeAddress2",
        "patAssigneeCity",
        "patAssigneeState",
        "patAssigneePostcode",
        "patAssigneeCountryName",
    ]
    output["assignees"] = zip_lists(output, assignee_fields, "assignee")
    # Collect properties into a list of dicts
    property_fields = [
        "inventionTitle",
        "inventionTitleLang",
        "applNum",
        "filingDate",
        "intlPublDate",
        "intlRegNum",
        "inventors",
        "issueDate",
        "patNum",
        "pctNum",
        "publDate",
        "publNum",
    ]
    output["properties"] = zip_lists(output, property_fields, "property")
    # Delete the original fields
    for key in assignor_fields + assignee_fields + property_fields:
        del output[key]
    # Collect the correspondent into a dict
    corr_address_fields = ["corrAddress1", "corrAddress2", "corrAddress3"]
    correspondent_address = "\n".join(output[k] for k in corr_address_fields if k in output)
    if correspondent_address:
        output["corr_address"] = correspondent_address
    for key in corr_address_fields:
        if key in output:
            del output[key]
    output["correspondent"] = {
        "name": output["corrName"],
        "address": output["corr_address"],
    }
    del output["corrName"]
    del output["corr_address"]

    # Collect the address for each assignee into a single string
    for assignee in output["assignees"]:
        address_lines = "\n".join(assignee[k] for k in ["patAssigneeAddress1", "patAssigneeAddress2"] if assignee[k])
        last_line = f"{assignee['patAssigneeCity']}, {assignee['patAssigneeState']} {assignee['patAssigneePostcode']}"
        if assignee["patAssigneeCountryName"]:
            last_line += f" ({assignee['patAssigneeCountryName']})"
        assignee_address = "\n".join([address_lines, last_line])
        if assignee_address:
            assignee["patAssigneeAddress"] = assignee_address
        for key in [
            "patAssigneeAddress1",
            "patAssigneeAddress2",
            "patAssigneeCity",
            "patAssigneeState",
            "patAssigneePostcode",
            "patAssigneeCountryName",
        ]:
            del assignee[key]

    return output


def zip_lists(data, input_keys, output_key):
    """Zip lists of data into a list of dicts"""
    tuples = list(zip(*[data[key] for key in input_keys]))
    dicts = [dict(zip(input_keys, t)) for t in tuples]
    return dicts


def convert_xml_to_json(xml_text) -> dict:
    """Convert the idiosyncratic xml of the Assignment API to ordinary json"""
    tree = ET.fromstring(xml_text)
    output = dict()
    num_found = tree.find(".//result")
    output["numFound"] = num_found.attrib["numFound"] if num_found is not None else 0
    output["docs"] = [convert_doc(doc) for doc in tree.findall(".//result/doc")]
    return output



from typing import List
from typing import Optional

from pydantic import Field

from ..schema.fulltext import ClaimsSchema
from ..schema.fulltext import DescriptionSchema
from patent_client.epo.ops.util import EpoBaseModel
from patent_client.util.claims.model import Claim


class FTDocumentId(EpoBaseModel):
    country: Optional[str] = None
    doc_number: Optional[str] = None
    kind: Optional[str] = None

    def __str__(self):
        return f"{self.country}{self.doc_number}{self.kind}"


class Claims(EpoBaseModel):
    __schema__ = ClaimsSchema()
    document_id: Optional[FTDocumentId] = None
    claims: List[Claim] = Field(default_factory=list)
    claim_text: Optional[str] = None

    @property
    def docdb_number(self) -> str:
        return str(self.document_id)

    def __repr__(self):
        return f"Claims(document_id={str(self.document_id)})"


class Description(EpoBaseModel):
    __schema__ = DescriptionSchema()
    document_id: Optional[FTDocumentId] = None
    description: Optional[str] = None

    @property
    def docdb_number(self) -> str:
        return str(self.document_id)



import math

from .api import PublishedAsyncApi
from .cql import generate_query
from .model.biblio import BiblioResult
from .model.fulltext import Claims
from .model.fulltext import Description
from .model.images import ImageDocument
from patent_client.util.manager import Manager


def get_ranges(limit=None, offset=0, page_size=100):
    offset = offset or 0
    offset += 1
    if not limit:
        page_no = 0
        while True:
            start = page_no * page_size + offset
            yield (start, start + page_size - 1)
            page_no += 1
    else:
        num_full_pages = math.floor(limit / page_size)
        last_page_size = limit % page_size
        for i in range(num_full_pages):
            start = i * page_size + offset
            yield (start, start + page_size - 1)
        if last_page_size:
            start = num_full_pages * page_size + offset
            yield (start, start + last_page_size - 1)


class SearchManager(Manager["BiblioResult"]):
    result_size = 100
    primary_key = "publication"

    async def _aget_search_results_range(self, start=1, end=100):
        if "cql_query" in self.config.filter:
            query = self.config.filter["cql_query"]
        else:
            query = generate_query(**self.config.filter)
        return await PublishedAsyncApi.search.search(query, start, end)

    async def alen(self) -> int:
        page = await self._aget_search_results_range(1, 100)
        offset = self.config.offset or 0
        limit = self.config.limit or page.num_results - offset
        num_results = page.num_results
        num_results -= offset
        num_results = min(limit, num_results)
        return num_results

    async def _aget_results(self):

        for start, end in get_ranges(self.config.limit, self.config.offset, self.result_size):
            page = await self._aget_search_results_range(start, end)
            for result in page.results:
                yield result
            if len(page.results) < self.result_size:
                break

    async def aget(self, number, doc_type="publication", format="docdb") -> BiblioResult:
        result = await PublishedAsyncApi.biblio.get_biblio(number, doc_type, format)
        if len(result.documents) > 1:
            raise Exception("More than one result found! Try another query")
        return result.documents[0]


class BiblioManager(Manager):
    async def aget(self, doc_number) -> "BiblioResult":
        result = await PublishedAsyncApi.biblio.get_biblio(doc_number)
        if len(result.documents) > 1:
            raise ValueError(f"More than one result found for {doc_number}!")
        return result.documents[0]


class ClaimsManager(Manager):
    async def aget(self, doc_number) -> "Claims":
        return await PublishedAsyncApi.fulltext.get_claims(doc_number)


class DescriptionManager(Manager):
    async def aget(self, doc_number) -> Description:
        return await PublishedAsyncApi.fulltext.get_description(doc_number)


class ImagesManager(Manager):
    async def aget(self, doc_number) -> ImageDocument:
        return await PublishedAsyncApi.images.get_images(doc_number)


class InpadocManager(SearchManager):
    pass


class InpadocBiblioManager(BiblioManager):
    pass



from yankee.json.schema import fields as f
from yankee.json.schema import Schema

from .shared import DocumentStructureSchema


class PublicSearchBiblioSchema(Schema):
    guid = f.String("guid")

    appl_id = f.String("applicationNumber")
    app_filing_date = f.Date("applicationFilingDate.0")
    related_appl_filing_date = f.List(f.Date, "relatedApplFilingDate")
    publication_number = f.String("publicationReferenceDocumentNumber")
    kind_code = f.String("kindCode.0")
    publication_date = f.Date("datePublished")
    patent_title = f.String("inventionTitle")

    inventors_short = f.String("inventorsShort")
    applicant_name = f.List(f.String, "applicantName")
    assignee_name = f.List(f.String, "assigneeName")
    government_interest = f.List(f.String, "governmentInterest")
    primary_examiner = f.String("primaryExaminer")
    assistant_examiner = f.List(f.String, "assistantExaminer")

    main_classification_code = f.String("mainClassificationCode")
    cpc_additional = f.DelimitedString(f.Str(), "cpcAdditionalFlattened", delimeter=";")
    cpc_inventive = f.DelimitedString(f.Str(), "cpcInventiveFlattened", delimeter=";")
    ipc_code = f.DelimitedString(f.Str(), "ipcCodeFlattened", delimeter=";")
    uspc_full_classification = f.DelimitedString(f.Str(), "uspcFullClassificationFlattened", delimeter=";")

    image_file_name = f.String("imageFileName")
    image_location = f.String("imageLocation")
    document_structure = DocumentStructureSchema(data_key=False)

    type = f.String("type")
    database_name = f.String("databaseName")
    composite_id = f.String("compositeId")
    document_id = f.String("documentId")
    document_size = f.Integer("documentSize")
    family_identifier_cur = f.Integer("familyIdentifierCur")
    language_indicator = f.String("languageIndicator")

    score = f.Float("score")


class PublicSearchBiblioPageSchema(Schema):
    num_found = f.Integer("numFound")
    per_page = f.Integer("perPage")
    page = f.Integer("page")
    docs = f.List(PublicSearchBiblioSchema, "patents")



from .model import Patent
from .model import PatentBiblio
from .model import PublicSearchBiblio
from .model import PublicSearchDocument
from .model import PublishedApplication
from .model import PublishedApplicationBiblio



import datetime

import pytest

from .util import date_to_solr_date


class TestDateToSolrDate:
    def test_can_take_string_date(self):
        assert date_to_solr_date("2021-01-01") == "2021-01-01T00:00:00Z"

    def test_can_take_datetime_date(self):
        assert date_to_solr_date(datetime.date(2021, 1, 1)) == "2021-01-01T00:00:00Z"

    def test_can_take_datetime(self):
        assert date_to_solr_date(datetime.datetime(2021, 1, 1, 12, 30, 0)) == "2021-01-01T12:30:00Z"

    def test_rejects_invalid_input(self):
        with pytest.raises(ValueError):
            date_to_solr_date(False)



import json
from pathlib import Path

import pytest

from .model import Family
from patent_client.util.test import compare_dicts

fixtures = Path(__file__).parent / "fixtures"


def test_model():
    result = Family.objects.get("EP1000000A1")
    expected_file = fixtures / "family_model_output.json"
    # expected_file.write_text(result.model_dump_json(indent=2))
    expected = json.loads(expected_file.read_text())
    compare_dicts(json.loads(result.model_dump_json()), expected)


@pytest.mark.asyncio
async def test_async_model():
    result = await Family.objects.aget("EP1000000A1")
    expected_file = fixtures / "family_model_output.json"
    # expected_file.write_text(result.model_dump_json(indent=2))
    expected = json.loads(expected_file.read_text())
    compare_dicts(json.loads(result.model_dump_json()), expected)



import datetime
from typing import Optional

from pydantic import AliasPath
from pydantic import Field
from pydantic import model_validator

from ..convert.biblio import PublicSearchBiblioPageSchema
from .shared import ApplicationNumber
from .shared import DateTimeAsDate
from .shared import DocumentStructure
from .shared import OptionalList
from patent_client.util.pydantic_util import BaseModel
from patent_client.util.related import get_model


class PublicSearchBiblio(BaseModel):
    guid: Optional[str] = None
    publication_number: Optional[str] = None
    publication_date: datetime.date
    patent_title: str
    type: str
    main_classification_code: str
    applicant_names: OptionalList[str] = Field(alias="applicant_name", default_factory=list)
    assignee_names: OptionalList[str] = Field(alias="assignee_name", default_factory=list)
    uspc_full_classification: OptionalList[str] = Field(default_factory=list)
    ipc_code: OptionalList[str] = Field(default_factory=list)
    cpc_additional: OptionalList[str] = Field(default_factory=list)
    app_filing_date: datetime.date
    related_appl_filing_date: OptionalList[DateTimeAsDate] = Field(alias="relatedApplFilingDate", default_factory=list)
    primary_examiner: Optional[str] = None
    assistant_examiner: Optional[str] = Field(alias=AliasPath("assistant_examiner", 0), default=None)
    appl_id: ApplicationNumber
    document_structure: DocumentStructure

    def __repr__(self):
        return f"PublicSearcBiblio(publication_number={self.publication_number}, publication_date={self.publication_date}, patent_title={self.patent_title})"

    @property
    def document(self):
        return get_model("patent_client.uspto.public_search.model.PatentDocument").objects.get(guid=self.guid)

    @property
    def forward_citations(self):
        return get_model("patent_client.uspto.public_search.model.PublicSearch").objects.filter(
            us_reference=self.publication_number
        )

    @property
    def application(self):
        return get_model("patent_client.uspto.peds.model.USApplication").objects.get(self.appl_id)

    @property
    def global_dossier(self):
        return get_model("patent_client.uspto.global_dossier.model.GlobalDossierApplication").objects.get(self.appl_id)

    @property
    def assignments(self):
        return get_model("patent_client.uspto.assignment.model.Assignment").objects.filter(appl_id=self.appl_id)

    @property
    def inpadoc(self):
        return get_model("patent_client.epo.ops.published.model.Inpadoc").objects.get("US" + self.publication_number)

    def download_images(self, path="."):
        return run_sync(self.adownload_images(path))

    async def adownload_images(self, path="."):
        from .manager import public_search_api

        return await public_search_api.download_image(self, path)


class PublicSearchBiblioPage(BaseModel):
    __schema__ = PublicSearchBiblioPageSchema()
    num_found: int
    per_page: int
    page: int
    docs: OptionalList[PublicSearchBiblio] = Field(default_factory=list)

    @model_validator(mode="before")
    @classmethod
    def run_convert(cls, values):
        return cls.__schema__.deserialize(values).to_dict()



import datetime
from pathlib import Path
from typing import *

from pydantic import BeforeValidator
from pydantic import ConfigDict
from pydantic import Field
from pydantic import model_validator
from pydantic.alias_generators import to_camel
from typing_extensions import Annotated

from patent_client.util.asyncio_util import run_sync
from patent_client.util.pydantic_util import BaseModel
from patent_client.util.related import get_model

MDYDate = Annotated[datetime.date, BeforeValidator(lambda x: datetime.datetime.strptime(x, "%m/%d/%Y").date())]
OptionalStrList = Annotated[List[str], BeforeValidator(lambda x: x if isinstance(x, list) else list())]


class GlobalDossierBaseModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        str_strip_whitespace=True,
    )


class GlobalDossierPublication(GlobalDossierBaseModel):
    pub_country: Optional[str] = None
    pub_num: Optional[str] = None
    kind_code: Optional[str] = None
    pub_date: Optional[MDYDate] = Field(alias="pubDateStr")


class GlobalDossierPriorityClaim(GlobalDossierBaseModel):
    country: Optional[str] = None
    doc_number: Optional[str] = None
    kind_code: Optional[str] = None


class GlobalDossierDocumentNumber(GlobalDossierBaseModel):
    country: Optional[str] = None
    doc_number: Optional[str] = None
    format: Optional[str] = None
    date: Optional[MDYDate] = None
    kind_code: Optional[str] = None


class GlobalDossierApplication(GlobalDossierBaseModel):
    app_num: Optional[str] = None
    app_date: Optional[MDYDate] = Field(alias="appDateStr")
    country_code: Optional[str] = None
    kind_code: Optional[str] = None
    doc_num: Optional["GlobalDossierDocumentNumber"] = None
    title: Optional[str] = None
    applicant_names: OptionalStrList = Field(default_factory=list)
    ip_5: Optional[bool] = Field(alias="ip5")
    priority_claim_list: "List[GlobalDossierPriorityClaim]" = Field(default_factory=list)
    pub_list: "List[GlobalDossierPublication]" = Field(default_factory=list)

    def __repr__(self):
        return f"GlobalDossierApplication(app_num={self.app_num}, country_code={self.country_code})"

    @property
    def document_list(self):
        return get_model("patent_client.uspto.global_dossier.model.DocumentList").objects.get(
            self.country_code, self.app_num, self.kind_code
        )

    @property
    def documents(self):
        return (
            get_model("patent_client.uspto.global_dossier.model.DocumentList")
            .objects.get(self.country_code, self.app_num, self.kind_code)
            .docs
        )

    @property
    def office_actions(self):
        return (
            get_model("patent_client.uspto.global_dossier.model.DocumentList")
            .objects.get(self.country_code, self.app_num, self.kind_code)
            .office_action_docs
        )

    @property
    def us_application(self):
        if self.country_code != "US":
            raise ValueError(f"Global Dossier Application is not a US Application! {self}")
        return get_model("patent_client.uspto.peds.model.USApplication").objects.get(self.app_num)

    @property
    def us_publication(self):
        if self.country_code != "US":
            raise ValueError(f"Global Dossier Application is not a US Application! {self}")
        pub = list(p for p in self.pub_list if p.kind_code == "A1")
        if len(pub) > 1:
            raise ValueError("More than one US publication for application!")
        return get_model("patent_client.uspto.public_search.model.PublishedApplication").objects.get(pub[0].pub_num)

    @property
    def us_patent(self):
        if self.country_code != "US":
            raise ValueError(f"Global Dossier Application is not a US Application! {self}")
        pat = list(p for p in self.pub_list if p.kind_code in ("A", "B1", "B2"))
        if len(pat) > 1:
            raise ValueError("More than one US patent for application!")
        return get_model("patent_client.uspto.public_search.model.Patent").objects.get(pat[0].pub_num)

    @property
    def us_assignments(self):
        if self.country_code != "US":
            raise ValueError(f"Global Dossier Application is not a US Application! {self}")
        return get_model("patent_client.uspto.assignment.model.Assignment").objects.filter(appl_id=self.app_num)


TextBool = Annotated[bool, BeforeValidator(lambda x: x == "true")]


class GlobalDossier(GlobalDossierBaseModel):
    country: Optional[str] = None
    internal: Optional[TextBool] = None
    corr_app_num: Optional[str] = None
    id: Optional[str] = None
    type: Optional[str] = None
    applications: List[GlobalDossierApplication] = Field(default_factory=list, alias="list")

    def __repr__(self):
        return f"GlobalDossier(id={self.id}, type={self.type}, country={self.country})"


class Document(GlobalDossierBaseModel):
    doc_number: Optional[str] = None
    country: Optional[str] = None
    doc_code: Optional[str] = None
    doc_desc: Optional[str] = None
    doc_id: Optional[str] = None
    date: Optional[MDYDate] = Field(alias="legalDateStr")
    doc_format: Optional[str] = None
    pages: Optional[int] = Field(alias="numberOfPages")
    doc_code_desc: Optional[str] = None
    doc_group_code: Optional[str] = None
    shareable: Optional[bool] = None

    def download(self, filename="", path="."):
        return run_sync(self.adownload(filename, path))

    async def adownload(self, filename="", path="."):
        from .manager import global_dossier_api

        out_path = Path(path).expanduser() / (
            f'{self.date.isoformat()} - {self.doc_desc.replace("/", "_")}.pdf' if not filename else filename
        )
        return await global_dossier_api.get_document(self.country, self.doc_number, self.doc_id, out_path=out_path)


class DocumentList(GlobalDossierBaseModel):
    title: Optional[str] = None
    doc_number: Optional[str] = None
    country: Optional[str] = None
    message: Optional[str] = None
    applicant_names: "List[str]" = Field(default_factory=list)
    office_action_count: Optional[int] = Field(alias="oaIndCount")
    docs: List[Document] = Field(default_factory=list)
    office_action_docs: List[Document] = Field(default_factory=list)

    @model_validator(mode="before")
    @classmethod
    def annotate_docs_and_office_actions(cls, values):
        update_dict = {
            "country": values["country"],
            "docNumber": values["docNumber"],
        }
        for doc in values["docs"]:
            doc.update(update_dict)
        for doc in values["officeActionDocs"]:
            doc.update(update_dict)
        return values



from .biblio import PublicSearchBiblio
from .biblio import PublicSearchBiblioPage
from .document import PublicSearchDocument


class PatentBiblio(PublicSearchBiblio):
    pass


class Patent(PublicSearchDocument):
    pass


class PublishedApplicationBiblio(PublicSearchBiblio):
    pass


class PublishedApplication(PublicSearchDocument):
    pass



import datetime
import re
from pathlib import Path
from typing import List
from typing import Optional
from typing import TYPE_CHECKING

from pydantic import BeforeValidator
from pydantic import ConfigDict
from pydantic import Field
from pydantic.alias_generators import to_camel
from typing_extensions import Annotated

from ...util.related import get_model
from .session import session
from patent_client.util.asyncio_util import run_sync
from patent_client.util.pydantic_util import BaseModel
from patent_client.util.pydantic_util import DateTime

if TYPE_CHECKING:
    from patent_client.uspto.peds.model import USApplication

MDYDate = Annotated[datetime.date, BeforeValidator(lambda x: datetime.datetime.strptime(x, "%m-%d-%Y").date())]


class PtabBaseModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        str_strip_whitespace=True,
    )


class AdditionalRespondent(PtabBaseModel):
    application_number_text: Optional[str] = None
    inventor_name: Optional[str] = None
    patent_number: Optional[str] = None
    party_name: Optional[str] = None


class PtabProceeding(PtabBaseModel):
    """A PTAB Proceeding - e.g. IPR/CBM/DER Trial, Patent Appeal, Interference, etc.

    All fields are query-able. Date ranges can be formed by inserting "from" or "to" on a query
    for a date range.

    """

    # Proceeding Metadata
    last_modified_date: Optional[DateTime] = None
    last_modified_user_id: Optional[DateTime] = None
    institution_decision_date: Optional[MDYDate] = None
    proceeding_filing_date: Optional[MDYDate] = None
    accorded_filing_date: Optional[MDYDate] = None
    proceeding_status_category: Optional[str] = None
    proceeding_number: Optional[str] = None
    proceeding_last_modified_date: Optional[MDYDate] = None
    proceeding_type_category: Optional[str] = None
    subproceeding_type_category: Optional[str] = None
    decision_date: Optional[MDYDate] = None
    docket_notice_mail_date: Optional[MDYDate] = None
    declaration_date: Optional[MDYDate] = None
    style_name_text: Optional[str] = None

    # Respondent Information
    respondent_technology_center_number: Optional[str] = None
    respondent_patent_owner_name: Optional[str] = None
    respondent_party_name: Optional[str] = None
    respondent_group_art_unit_number: Optional[str] = None
    respondent_inventor_name: Optional[str] = None
    respondent_counsel_name: Optional[str] = None
    respondent_grant_date: Optional[MDYDate] = None
    respondent_patent_number: Optional[str] = None
    respondent_application_number_text: Optional[str] = None
    respondent_publication_number: Optional[str] = None
    respondent_publication_date: Optional[MDYDate] = None

    # Petitioner Information
    petitioner_technology_center_number: Optional[str] = None
    petitioner_patent_owner_name: Optional[str] = None
    petitioner_party_name: Optional[str] = None
    petitioner_group_art_unit_number: Optional[str] = None
    petitioner_inventor_name: Optional[str] = None
    petitioner_counsel_name: Optional[str] = None
    petitioner_grant_date: Optional[MDYDate] = None
    petitioner_patent_number: Optional[str] = None
    petitioner_application_number_text: Optional[str] = None

    # Appellant Information
    appellant_technology_center_number: Optional[str] = None
    appellant_patent_owner_name: Optional[str] = None
    appellant_party_name: Optional[str] = None
    appellant_group_art_unit_number: Optional[str] = None
    appellant_inventor_name: Optional[str] = None
    appellant_counsel_name: Optional[str] = None
    appellant_grant_date: Optional[MDYDate] = None
    appellant_patent_number: Optional[str] = None
    appellant_application_number_text: Optional[str] = None
    appellant_publication_date: Optional[MDYDate] = None
    appellant_publication_number: Optional[str] = None
    third_party_name: Optional[str] = None

    # Second Respondent (if any)
    second_respondent_party_name: Optional[str] = None
    second_respondent_appl_number_text: Optional[str] = None
    second_respondent_patent_number: Optional[str] = None
    second_respondent_grant_date: Optional[MDYDate] = None
    second_respondent_patent_owner_name: Optional[str] = None
    second_respondent_inventor_name: Optional[str] = None
    second_respondent_counsel_name: Optional[str] = None
    second_respondent_g_a_u_number: Optional[str] = None
    second_respondent_tech_center_number: Optional[str] = None
    second_respondent_pub_number: Optional[str] = None
    second_respondent_publication_date: Optional[MDYDate] = None

    additional_respondents: List[str] = Field(default_factory=list)

    @property
    def documents(
        self,
    ) -> "list[PtabDocument]":
        """Documents associated with the Proceeding"""
        return get_model("patent_client.uspto.ptab.model.PtabDocument").objects.filter(
            proceeding_number=self.proceeding_number
        )

    @property
    def decisions(
        self,
    ) -> "list[PtabDecision]":
        """Decisions associated with the Proceeding"""
        return get_model("patent_client.uspto.ptab.model.PtabDecision").objects.filter(
            proceeding_number=self.proceeding_number
        )

    @property
    def us_application(
        self,
    ) -> "list[USApplication]":
        """The US Application provided by PEDS associated with the Proceeding"""
        return get_model("patent_client.uspto.peds.model.USApplication").objects.get(
            patent_number=self.respondent_patent_number
        )


fname_re = re.compile(r"[<>:\"/\|?*]")


class PtabDocument(PtabBaseModel):
    document_identifier: str = Field(repr=False, default=None)
    document_category: Optional[str] = None
    document_type_name: Optional[str] = None
    document_number: Optional[int] = None
    document_name: Optional[str] = None
    document_filing_date: Optional[MDYDate] = None
    proceeding_number: Optional[str] = None
    proceeding_type_category: Optional[str] = Field(repr=False, default=None)
    title: Optional[str] = None

    @property
    def proceeding(self) -> "PtabProceeding":
        """The PTAB proceeding associated with the document"""
        return get_model("patent_client.uspto.ptab.model.PtabProceeding").objects.get(
            proceeding_number=self.proceeding_number
        )

    def download(self, path: Optional[str | Path]) -> Path:
        return run_sync(self.adownload(path))

    async def adownload(self, path: Optional[str | Path]) -> Path:
        name, ext = self.document_name.rsplit(".", 1)
        name = name[:100] + "." + ext
        filename = f"[{str(self.document_number).rjust(4, '0')}] {self.document_filing_date.isoformat()} - {name}"
        filename = filename.encode(encoding="ascii", errors="ignore").decode("ascii")
        filename = fname_re.sub("", filename)
        out_path = Path(path) if path else Path.cwd()
        if out_path.is_dir():
            out_path = Path(path) / filename
        else:
            out_path = out_path
        donwload_url = f"https://developer.uspto.gov/ptab-api/documents/{self.document_identifier}/download"
        return await session.download(download_url, path=out_path)


class PtabDecision(PtabBaseModel):
    __manager__ = "patent_client.uspto.ptab.manager.PtabDecisionManager"
    proceeding_number: str
    board_rulings: List[str] = Field(default_factory=list)
    decision_type_category: Optional[str] = None
    document_identifier: Optional[str] = None
    document_name: Optional[str] = None
    identifier: Optional[str] = None
    subdecision_type_category: Optional[str] = None
    issue_type: List[str] = Field(default_factory=list)
    object_uu_id: Optional[str] = None
    petitioner_technology_center_number: Optional[str] = None

    @property
    def proceeding(self) -> "PtabProceeding":
        """The PTAB proceeding associated with the document"""
        return get_model("patent_client.uspto.ptab.model.PtabProceeding").objects.get(
            proceeding_number=self.proceeding_number
        )


class PtabDocumentPage(PtabBaseModel):
    num_found: Optional[int] = Field(alias="recordTotalQuantity")
    docs: List[PtabDocument] = Field(alias="results", default_factory=list)


class PtabProceedingPage(PtabBaseModel):
    num_found: Optional[int] = Field(alias="recordTotalQuantity")
    docs: List[PtabProceeding] = Field(alias="results", default_factory=list)


class PtabDecisionPage(PtabBaseModel):
    num_found: Optional[int] = Field(alias="recordTotalQuantity")
    docs: List[PtabDecision] = Field(alias="results", default_factory=list)



import re

import lxml.html as ETH

newline_re = re.compile(r"<br />\s*")
bad_break_re = re.compile(r"<br />\s+")


def html_to_text(html):
    html = newline_re.sub("\n\n", html)
    # html = bad_break_re.sub(" ", html)
    return "".join(ETH.fromstring(html).itertext())



import datetime
from pathlib import Path

from .convert import convert_xml_to_json
from .model import Assignment
from .model import AssignmentPage

test_dir = Path(__file__).parent / "fixtures"


def test_assignment_1():
    xml_file = test_dir / "assignment_1.xml"
    input_data = convert_xml_to_json(xml_file.read_bytes())
    page = AssignmentPage.model_validate(input_data)
    a = page.docs[0]
    assert a.id == "18247-405"
    assert a.conveyance_text == "NUNC PRO TUNC ASSIGNMENT"
    assert a.last_update_date <= datetime.date.fromisoformat("2008-01-22")
    assert a.transaction_date.isoformat() == "2019-07-11"
    assert a.recorded_date.isoformat() == "2006-09-14"
    assert a.correspondent.name == "JEFFREY H. INGERMAN"
    assert (
        a.correspondent.address
        == "FISH & NEAVE IP GROUP, ROPES & GRAY LLP\n1251 AVENUE OF THE AMERICAS C3\nNEW YORK, NY 10020-1105"
    )
    assert len(a.assignors) == 1
    assert a.assignors[0].name == "REALTIME DATA COMPRESSION SYSTEMS, INC."
    assert a.assignors[0].execution_date.isoformat() == "2006-09-14"
    assert len(a.assignees) == 1
    assert a.assignees[0].name == "REALTIME DATA LLC"
    assert a.assignees[0].address == "15 WEST 36TH STREET\nNEW YORK, NEW YORK 10018"
    assert len(a.properties) == 5
    assert a.properties[0].invention_title == "SYSTEM AND METHODS FOR ACCELERATED DATA STORAGE AND RETRIEVAL"
    assert a.properties[0].appl_num == "10628795"
    assert a.properties[0].filing_date.isoformat() == "2003-07-28"
    assert a.properties[0].intl_publ_date == None
    assert a.properties[0].intl_reg_num == None
    assert a.properties[0].inventors == "James J. Fallon"
    assert a.properties[0].issue_date.isoformat() == "2006-10-31"
    assert a.properties[0].pat_num == "7130913"
    assert a.properties[0].pct_num == None
    assert a.properties[0].publ_date.isoformat() == "2004-04-15"
    assert a.properties[0].publ_num == "20040073746"


def test_manager():
    a = Assignment.objects.get("18247-405")
    assert a.id == "18247-405"



from dataclasses import dataclass
from dataclasses import fields


@dataclass
class QueryFields:
    app_type_facet: str = "appTypeFacet"
    appl_id: str = "applId"
    total_pto_days: str = "totalPtoDays"
    app_filing_date: str = "appFilingDate"
    first_named_applicant_address: str = "firstNamedApplicantAddress"
    app_location_facet: str = "appLocationFacet"
    app_location_date: str = "appLocationDate"
    inventor_name: str = "inventorName"
    app_early_pub_date: str = "appEarlyPubDate"
    intl_filing_year: str = "intlFilingYear"
    app_exam_mdl_name: str = "appExamMdlName"
    pta_pte_type: str = "ptaPteType"
    corr_addr_name_line_two: str = "corrAddrNameLineTwo"
    app_location: str = "appLocation"
    app_early_pub_number: str = "appEarlyPubNumber"
    wipo_early_pub_date: str = "wipoEarlyPubDate"
    id: str = "id"
    text: str = "text"
    app_trans_history_date: str = "appTransHistoryDate"
    primary_inventor_city: str = "primaryInventorCity"
    app_exam_last_name: str = "appExamLastName"
    first_named_applicant_name_list: str = "firstNamedApplicantNameList"
    app_pubyear: str = "appPubyear"
    app_sub_cls: str = "appSubCls"
    primary_inventor_country: str = "primaryInventorCountry"
    patent_number_facet: str = "patentNumberFacet"
    app_status: str = "appStatus"
    corr_addr_street_line_two: str = "corrAddrStreetLineTwo"
    pta_pte_ind: str = "ptaPteInd"
    corr_addr_geo_region_na: str = "corrAddrGeoRegionNA"
    first_named_applicant_address_list: str = "firstNamedApplicantAddressList"
    first_inventor_file_facet: str = "firstInventorFileFacet"
    c_delay: str = "cDelay"
    app_status_date: str = "appStatusDate"
    app_control_number: str = "appControlNumber"
    pto_adjustments: str = "ptoAdjustments"
    first_inventor_file: str = "firstInventorFile"
    app_grp_art_number_facet: str = "appGrpArtNumberFacet"
    overlap_delay: str = "overlapDelay"
    app_type: str = "appType"
    rank_and_inventors_list: str = "rankAndInventorsList"
    app_status_year: str = "appStatusYear"
    primary_inventor_middle_name: str = "primaryInventorMiddleName"
    pct_app_type: str = "pctAppType"
    app_cust_number: str = "appCustNumber"
    last_updated_timestamp: str = "lastUpdatedTimestamp"
    first_named_applicant_facet: str = "firstNamedApplicantFacet"
    wipo_early_pub_number: str = "wipoEarlyPubNumber"
    app_cls_sub_cls: str = "appClsSubCls"
    b_delay: str = "bDelay"
    corr_addr_name_line_one: str = "corrAddrNameLineOne"
    first_named_applicant: str = "firstNamedApplicant"
    app_exam_first_name: str = "appExamFirstName"
    patent_number_str: str = "patentNumberStr"
    app_pct_number: str = "appPCTNumber"
    corr_addr_country_cd: str = "corrAddrCountryCd"
    pto_delay: str = "ptoDelay"
    app_trans_history_desc: str = "appTransHistoryDesc"
    app_entity_status: str = "appEntityStatus"
    app_location_year: str = "appLocationYear"
    corr_addr_country_name: str = "corrAddrCountryName"
    app_early_pub_year: str = "appEarlyPubYear"
    app_exam_name: str = "appExamName"
    public_ind: str = "publicInd"
    app_cls_sub_cls_facet: str = "appClsSubClsFacet"
    app_exam_name_facet: str = "appExamNameFacet"
    primary_inventor_first_name: str = "primaryInventorFirstName"
    pct_app_type_facet: str = "pctAppTypeFacet"
    intl_filing_date: str = "intlFilingDate"
    primary_inventor_facet: str = "primaryInventorFacet"
    app_entity_status_txt: str = "appEntityStatus_txt"
    app_grp_art_number: str = "appGrpArtNumber"
    patent_kind_code: str = "patentKindCode"
    app_filing_year: str = "appFilingYear"
    app_attr_dock_number_facet: str = "appAttrDockNumberFacet"
    primary_inventor_region: str = "primaryInventorRegion"
    first_named_applicant_name: str = "firstNamedApplicantName"
    appl_id_txt: str = "appl_id_txt"
    patent_term_json: str = "patentTermJson"
    patent_number: str = "patentNumber"
    last_mod_ts_utc: str = "LAST_MOD_TS_UTC"
    app_cls: str = "appCls"
    last_insert_time: str = "LAST_INSERT_TIME"
    patent_issue_year: str = "patentIssueYear"
    app_status_txt: str = "appStatus_txt"
    app_cust_number_facet: str = "appCustNumberFacet"
    app_intl_pub_date: str = "appIntlPubDate"
    app_early_pub_number_facet: str = "appEarlyPubNumberFacet"
    patent_title: str = "patentTitle"
    appl_delay: str = "applDelay"
    appl_id_str: str = "applIdStr"
    app_pct_number_facet: str = "appPCTNumberFacet"
    app_attr_dock_number: str = "appAttrDockNumber"
    primary_inventor: str = "primaryInventor"
    app_pub_seq_number: str = "appPubSeqNumber"
    corr_addr_street_line_one: str = "corrAddrStreetLineOne"
    ip_office_code: str = "ipOfficeCode"
    a_delay: str = "aDelay"
    rank_and_inventors_list_str: str = "rankAndInventorsList_str"
    app_intl_pub_number_facet: str = "appIntlPubNumberFacet"
    publish_doc_json: str = "publishDocJson"
    rank_and_inventors_address_list: str = "rankAndInventorsAddressList"
    corr_addr_postal_code: str = "corrAddrPostalCode"
    app_intl_pub_year: str = "appIntlPubYear"
    primary_inventor_last_name: str = "primaryInventorLastName"
    patent_issue_date: str = "patentIssueDate"
    app_intl_pub_number: str = "appIntlPubNumber"
    app_control_number_facet: str = "appControlNumberFacet"
    corr_addr_cust_no: str = "corrAddrCustNo"
    corr_addr_city: str = "corrAddrCity"
    app_exam_prefrd_name: str = "appExamPrefrdName"
    app_confr_number: str = "appConfrNumber"

    @classmethod
    def field_names(cls):
        return [f.name for f in fields(cls)]

    @classmethod
    def is_date_field(cls, field_name):
        return any(field_name in f for f in cls._date_fields)

    @classmethod
    def get(cls, key):
        return getattr(cls, key)

    _date_fields = [
        "appFilingDate",
        "appLocationDate",
        "appEarlyPubDate",
        "wipoEarlyPubDate",
        "appTransHistoryDate",
        "appStatusDate",
        "lastUpdatedTimestamp",
        "intlFilingDate",
        "LAST_MOD_TS_UTC",
        "LAST_INSERT_TIME",
        "appIntlPubDate",
        "patentIssueDate",
    ]



import datetime
import warnings
from pathlib import Path
from typing import List
from typing import Optional
from typing import TYPE_CHECKING

from dateutil.parser import isoparse
from pydantic import BeforeValidator
from pydantic import ConfigDict
from pydantic import Field
from pydantic import field_validator
from pydantic.alias_generators import to_camel
from typing_extensions import Annotated

from patent_client.util.asyncio_util import run_sync
from patent_client.util.pydantic_util import BaseModel
from patent_client.util.related import get_model

if TYPE_CHECKING:
    from patent_client.uspto.peds.model import USApplication
    from patent_client.uspto.public_search.model import Patent, PublishedApplication
    from .manager import AssignmentManager

from .session import session

UsptoDate = Annotated[datetime.date, BeforeValidator(lambda x: datetime.datetime.strptime(x, "%Y%m%d").date())]
YNBool = Annotated[bool, BeforeValidator(lambda x: x == "Y")]


def parse_datetime(string):
    dt = isoparse(string)
    if dt.year == 1 and dt.month == 1 and dt.day == 1:
        return None
    return dt


def parse_date(string):
    dt = parse_datetime(string)
    if dt is None:
        return None
    return dt.date()


DatetimeAsDate = Annotated[Optional[datetime.date], BeforeValidator(lambda x: parse_date(x))]


class AbstractAssignmentModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        str_strip_whitespace=True,
    )


class Assignor(AbstractAssignmentModel):
    name: str = Field(alias="patAssignorName")
    execution_date: DatetimeAsDate = Field(alias="patAssignorExDate")
    acknowledgement_date: DatetimeAsDate = Field(alias="patAssignorDateAck")


class Assignee(AbstractAssignmentModel):
    name: str = Field(alias="patAssigneeName")
    address: str = Field(alias="patAssigneeAddress")


class Property(AbstractAssignmentModel):
    invention_title: str
    invention_title_lang: str
    appl_num: str
    filing_date: Optional[DatetimeAsDate] = None
    intl_publ_date: Optional[DatetimeAsDate] = None
    intl_reg_num: Optional[str] = None
    inventors: Optional[str] = None
    issue_date: Optional[DatetimeAsDate] = None
    pat_num: Optional[str] = None
    pct_num: Optional[str] = None
    publ_date: Optional[DatetimeAsDate] = None
    publ_num: Optional[str] = None

    @property
    def application(self) -> Optional["USApplication"]:
        """The related US Application"""
        try:
            appl_id = getattr(self, "appl_id", None)
            if appl_id is not None:
                return get_model("patent_client.uspto.peds.model.USApplication").objects.get(appl_id=appl_id)
            appl_id = getattr(self, "pct_num", None)
            if appl_id is not None:
                return get_model("patent_client.uspto.peds.model.USApplication").objects.get(appl_id=appl_id)
            pub_num = getattr(self, "publ_num", None)
            if pub_num is not None:
                return get_model("patent_client.uspto.peds.model.USApplication").objects.get(
                    app_early_pub_number=pub_num
                )
            pat_num = getattr(self, "pat_num", None)
            if pat_num is not None:
                return get_model("patent_client.uspto.peds.model.USApplication").objects.get(patent_number=pat_num)
        except Exception:
            pass
        warnings.warn(f"Unable to find application for {self}")
        return None

    @property
    def patent(self) -> "Patent":
        """The related US Patent, if any"""
        return get_model("patent_client.uspto.public_search.model.Patent").objects.get(publication_number=self.pat_num)

    @property
    def publication(
        self,
    ) -> "PublishedApplication":
        """The related US Publication, if any"""
        return get_model("patent_client.uspto.peds.fulltext.patent.model.PublishedApplication").objects.get(
            publication_number=self.publ_num
        )


class Correspondent(AbstractAssignmentModel):
    name: str
    address: str


class Assignment(AbstractAssignmentModel["AssignmentManager"]):
    id: str
    date_produced: Optional[UsptoDate] = None
    action_key_code: Optional[str] = None
    transaction_date: Optional[DatetimeAsDate] = None
    last_update_date: DatetimeAsDate
    purge_indicator: YNBool
    recorded_date: DatetimeAsDate
    page_count: int
    conveyance_text: str
    assignment_record_has_images: YNBool
    attorney_dock_num: Optional[str] = None
    pat_assignor_earliest_ex_date: DatetimeAsDate
    correspondent: Correspondent
    assignors: List[Assignor]
    assignees: List[Assignee]
    properties: List[Property]

    @field_validator("conveyance_text")
    @classmethod
    def remove_see_document(cls, v) -> str:
        return v.replace(" (SEE DOCUMENT FOR DETAILS).", "").strip()

    @property
    def reel_frame(self):
        return self.id.split("-")

    @property
    def image_url(self):
        reel, frame = self.reel_frame
        reel = reel.rjust(6, "0")
        frame = frame.rjust(4, "0")
        return f"http://legacy-assignments.uspto.gov/assignments/assignment-pat-{reel}-{frame}.pdf"

    def download(self, path: Optional[str | Path] = None):
        """downloads the PDF associated with the assignment to the current working directory"""
        return run_sync(self.adownload(path=path))

    async def adownload(self, path: Optional[str | Path] = None):
        """asynchronously downloads the PDF associated with the assignment to the current working directory"""
        return await session.download(self.image_url, path=path)


class AssignmentPage(AbstractAssignmentModel):
    num_found: int
    docs: list[Assignment]



from .model import NumberServiceResult
from patent_client.epo.ops.session import asession


class NumberServiceException(Exception):
    pass


class NumberServiceApi:
    @classmethod
    async def convert_number(
        cls, number, doc_type="publication", input_format="original", output_format="docdb"
    ) -> NumberServiceResult:
        """Number Conversion Service
        number: number to convert. Must include at least country and doc number,
            can optionally include kind code or publication date. e.g.
            MD.20050130
            MD.20050130.A
            MD.20050130.A.20050130
            JP.2006-147056.A.20060526
            JP.2006147056.A.20060526
            US.(08/921,321).A.19970829
        doc_type: type of document (publication / application / priority)
        input_format: input type (original / docdb / epodoc)
        output_format: output type (original / docdb / epodoc)

        """
        response = await asession.get(
            f"http://ops.epo.org/3.2/rest-services/number-service/{doc_type}/{input_format}/{number}/{output_format}"
        )
        result = NumberServiceResult.model_validate(response.text)
        errors = [m for m in result.messages if m["kind"] == "ERROR"]
        if errors:
            raise NumberServiceException("\n".join(str(e) for e in errors))
        return result



import datetime
import logging
import re
import sqlite3
from pathlib import Path

import lxml.etree as ET
from openpyxl import load_workbook

from patent_client import BASE_DIR
from patent_client.session import PatentClientSession
from patent_client.util.asyncio_util import run_sync

session = PatentClientSession()

legal_code_dir = BASE_DIR / "epo"
legal_code_dir.mkdir(exist_ok=True, parents=True)
db_location = legal_code_dir / "legal_codes.sqlite"


logger = logging.getLogger(__name__)


def current_date():
    return datetime.datetime.now().date()


def generate_legal_code_db():
    current = has_current_spreadsheet()
    if current:
        logger.info("Legal Code Database is Current - skipping database creation")
        return
    else:
        try:
            logger.info("Legal Code Database is out of date - creating legal code database")
            path = get_spreadsheet()
        except Exception:
            logger.exception("Could not find live code file - falling back to default dated 2023-11-05")
            path = Path(__file__).parent / "legal_code_descriptions_2023-44-0.xlsx"
        create_code_database(path)


def has_current_spreadsheet():
    con = sqlite3.connect(db_location, timeout=30)
    cur = con.cursor()
    try:
        fname = cur.execute("SELECT * FROM meta").fetchone()[0]
        date_string = re.search(r"legal_code_descriptions_([\d-]+)\.xlsx", fname).group(1)
        try:
            date = datetime.datetime.strptime(date_string, "%Y-%W-%w").date()
        except ValueError:  # Indicates using an older format and database needs to be updated
            return False
        age = datetime.datetime.now().date() - date
        logger.debug(f"Legal Code Database is {age} days old")
        return age.days < 7
    except (sqlite3.OperationalError, TypeError):
        return False


def get_spreadsheet_from_epo_website() -> tuple[datetime.date, str]:
    url = "https://www.epo.org/searching-for-patents/data/coverage/weekly.html"
    response = run_sync(session.get(url))
    response.raise_for_status()
    tree = ET.HTML(response.text)
    date_string = tree.xpath(".//tr/td[contains(text(), 'Legal event codes')]/../td[4]")[0].text.strip()
    week, year = date_string.split()[1].split("/")
    date = datetime.datetime.strptime(f"{year}-{week}-0", "%Y-%W-%w").date()
    excel_url = tree.xpath('.//a[contains(@href, "Legal-event-codes")][1]/@href')[0]
    logger.info(f"Found new spreadsheet for {date.isoformat()}: {excel_url}")
    return (date, excel_url)


def get_spreadsheet() -> tuple[datetime.date, Path]:
    date, excel_url = get_spreadsheet_from_epo_website()
    out_path = legal_code_dir / f"legal_code_descriptions_{date.strftime('%Y-%W-%w')}.xlsx"
    if out_path.exists():
        logger.info(f"File already downloaded! Current as of {date.isoformat()}")
        return out_path
    out_path = session.download(excel_url, path=out_path)
    logger.info(f"Downloaded new live code file for date {date.isoformat()}")
    return out_path


def create_code_database(excel_path):
    con = sqlite3.connect(db_location, timeout=30)
    cur = con.cursor()
    try:
        meta = cur.execute("SELECT * FROM meta").fetchone()[0]
        if meta == excel_path.name:
            logger.debug(f"Excel file {excel_path.name} already loaded. Skipping!")
            return
    except (sqlite3.OperationalError, TypeError):
        pass

    cur.execute("CREATE TABLE IF NOT EXISTS meta (file_name text)")
    cur.execute("INSERT INTO meta values (?)", (excel_path.name,))
    wb = load_workbook(excel_path)
    data = list(tuple(i.strip() for i in r) for r in wb[wb.sheetnames[0]].iter_rows(values_only=True))
    rows = data[1:]
    cur.execute(
        """CREATE TABLE IF NOT EXISTS legal_codes (
    country_code text,
    event_code text,
    date_created text,
    influence text,
    description text,
    last_update text,
    description_orig text,
    last_update_orig text,
    event_class text,
    event_class_description text)"""
    )
    cur.execute("""CREATE INDEX IF NOT EXISTS country_event_code ON legal_codes (country_code, event_code)""")
    cur.executemany("INSERT INTO legal_codes values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", rows)
    con.commit()


class LegalCodes:
    def __init__(self):
        self.connection = sqlite3.connect(db_location, timeout=30)
        self.connection.row_factory = sqlite3.Row

    def get_code_data(self, country_code, legal_code):
        cur = self.connection.cursor()
        try:
            return dict(
                cur.execute(
                    "SELECT * FROM legal_codes WHERE country_code = ? AND event_code = ?",
                    (country_code, legal_code),
                ).fetchone()
            )
        except TypeError:
            raise Exception(f"No Event Data found for {country_code} - {legal_code}")



from typing import List
from typing import Optional

from pydantic import Field

from .schema import FamilySchema
from patent_client.epo.ops.number_service.model import DocumentId
from patent_client.epo.ops.util import EpoBaseModel


class PriorityClaim(EpoBaseModel):
    application_number: Optional[str] = None
    application_reference: Optional[DocumentId] = None
    sequence: Optional[int] = None
    kind: Optional[str] = None
    active: Optional[bool] = None


class FamilyMember(EpoBaseModel):
    publication_number: Optional[str] = None
    application_number: Optional[str] = None
    family_id: Optional[str] = None
    publication_reference: list = Field(default_factory=list)
    application_reference: list = Field(default_factory=list)
    priority_claims: List[PriorityClaim] = Field(default_factory=list)

    @property
    def docdb_number(self):
        return self.publication_number


class Family(EpoBaseModel):
    __schema__ = FamilySchema()
    publication_reference: Optional[DocumentId] = None
    num_records: Optional[int] = None
    publication_number: Optional[str] = None
    family_members: List[FamilyMember] = Field(default_factory=list)



from .api import FamilyAsyncApi
from .model import Family
from .schema import FamilySchema
from patent_client.util.manager import Manager


class FamilyManager(Manager[Family]):
    __schema__ = FamilySchema

    async def aget(self, doc_number):
        return await FamilyAsyncApi.get_family(doc_number, doc_type="publication", format="docdb")



import json
from pathlib import Path

import pytest

from .api import FamilyAsyncApi

fixtures = Path(__file__).parent / "fixtures"


@pytest.mark.asyncio
async def test_api():
    result = await FamilyAsyncApi.get_family("EP1000000A1")
    expected_file = fixtures / "family_api_output.json"
    # expected_file.write_text(result.model_dump_json(indent=2))
    expected = expected_file.read_text()
    assert json.loads(expected) == json.loads(result.model_dump_json())



from pathlib import Path
from typing import List
from typing import Optional

from pydantic import computed_field
from pydantic import Field
from pypdf import PdfReader
from pypdf import PdfWriter

from ..schema.images import ImagesSchema
from patent_client.epo.ops.util import EpoBaseModel
from patent_client.epo.ops.util import InpadocModel


class Section(EpoBaseModel):
    name: Optional[str] = None
    start_page: Optional[int] = None


class ImageDocument(EpoBaseModel):
    num_pages: Optional[int] = None
    description: Optional[str] = None
    link: Optional[str] = None
    formats: List[str] = Field(default_factory=list)
    sections: List[Section] = Field(default_factory=list)
    doc_number: Optional[str] = None

    def download(self, path="."):
        from ..api import PublishedImagesApi

        out_file = Path(path) / f"{self.doc_number}.pdf"
        writer = PdfWriter()
        for i in range(1, self.num_pages + 1):
            page_data = PublishedImagesApi.get_page_image_from_link(self.link, page_number=i)
            page = PdfReader(page_data).pages[0]
            if page["/Rotate"] == 90:
                page.rotate_clockwise(-90)
            writer.add_page(page)

        for section in self.sections:
            writer.add_outline_item(section.name.capitalize(), section.start_page)

        with out_file.open("wb") as f:
            writer.write(f)

    def download_image(self, path=".", image_format="tif", page_number=1):
        from ..api import PublishedImagesApi

        out_file = Path(path) / f"{self.doc_number}.{image_format}"

        image = PublishedImagesApi.get_page_image_from_link(
            self.link, page_number=page_number, image_format=image_format
        )

        with out_file.open("wb") as f:
            f.write(image.read())


class Images(InpadocModel):
    __schema__ = ImagesSchema()
    publication_number: Optional[str] = None
    full_document: Optional[ImageDocument] = None
    drawing: Optional[ImageDocument] = None
    first_page: Optional[ImageDocument] = None

    @computed_field
    @property
    def docdb_number(self) -> str:
        return str(self.publication_number)



import datetime
import importlib
from typing import Generic
from typing import Optional
from typing import TypeVar

from dateutil.parser import isoparse
from dateutil.parser import parse as parse_dt
from pydantic import BaseModel as PydanticBaseModel
from pydantic import BeforeValidator
from pydantic import ConfigDict
from typing_extensions import Annotated

from patent_client.util.manager import Manager


def parse_datetime(date_obj: str | datetime.datetime) -> datetime.datetime:
    if isinstance(date_obj, datetime.datetime):
        return date_obj
    try:
        return isoparse(date_obj)
    except ValueError:
        return parse_dt(date_obj)


def parse_date(date_obj: str | datetime.datetime | datetime.date) -> datetime.date:
    if isinstance(date_obj, datetime.date):
        return date_obj
    elif isinstance(date_obj, datetime.datetime):
        return date_obj.date()
    try:
        return isoparse(date_obj).date()
    except ValueError:
        return parse_dt(date_obj).date()


DateTime = Annotated[datetime.datetime, BeforeValidator(parse_datetime)]
Date = Annotated[datetime.date, BeforeValidator(parse_date)]


class ClassProperty:
    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


class UnmanagedModelException(Exception):
    pass


M = TypeVar("M", bound=Manager)


class BaseModel(PydanticBaseModel, Generic[M]):
    __manager__: Optional[str] = None
    model_config = ConfigDict(
        ignored_types=(ClassProperty,),
    )

    @ClassProperty
    def objects(cls) -> M:
        if cls.__manager__ is not None:
            manager_module, manager_class_name = cls.__manager__.rsplit(".", 1)
            try:
                manager_class = getattr(importlib.import_module(manager_module), manager_class_name)
                return manager_class()
            except AttributeError as e:
                raise ValueError(
                    f"Specified manager at {cls.__manager__} not found! looked in {manager_module}.{manager_class_name}"
                )
        else:
            manager_module = cls.__module__.split(".model")[0] + ".manager"
            manager_class_name = cls.__name__ + "Manager"
        try:
            manager_class = getattr(importlib.import_module(manager_module), manager_class_name)
            return manager_class()
        except AttributeError as e:
            raise UnmanagedModelException(
                f"Unable to find manager for {cls.__name__}, looked in {manager_module}.{manager_class_name}"
            )

    def to_dict(self):
        return self.model_dump()

    def items(self):
        return iter(self)



import json
from pathlib import Path

from .convert import convert_xml_to_json

fixtures = Path(__file__).parent / "fixtures"


def test_assignment_1():
    input_doc = fixtures / "assignment_1.xml"
    output_doc = fixtures / "assignment_1.json"
    output_data = convert_xml_to_json(input_doc.read_bytes())
    # output_doc.write_text(json.dumps(output_data, indent=2))
    expected_data = json.loads(output_doc.read_text())
    assert output_data == expected_data


def test_assignment_2():
    input_doc = fixtures / "assignment_2.xml"
    output_doc = fixtures / "assignment_2.json"
    output_data = convert_xml_to_json(input_doc.read_bytes())
    # output_doc.write_text(json.dumps(output_data, indent=2))
    expected_data = json.loads(output_doc.read_text())
    assert output_data == expected_data


def test_assignment_3():
    input_doc = fixtures / "assignment_3.xml"
    output_doc = fixtures / "assignment_3.json"
    output_data = convert_xml_to_json(input_doc.read_bytes())
    # output_doc.write_text(json.dumps(output_data, indent=2))
    expected_data = json.loads(output_doc.read_text())
    assert output_data == expected_data



import datetime

from dateutil.parser import parse as dt_parse


def date_to_solr_date(date):
    if isinstance(date, datetime.datetime):
        pass
    elif isinstance(date, str):
        date = dt_parse(date)
    elif isinstance(date, datetime.date):
        date = datetime.datetime.combine(date, datetime.datetime.min.time())
    elif not isinstance(date, datetime.datetime):
        raise ValueError(f"Invalid date type: {type(date)}")
    return date.strftime("%Y-%m-%dT%H:%M:%SZ")



import json
from pathlib import Path

from .model.biblio import PublicSearchBiblioPage
from .model.document import PublicSearchDocument
from patent_client.util.test import compare_dicts
from patent_client.util.test import compare_lists

fixtures = Path(__file__).parent / "fixtures"


import datetime


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


def test_parse_biblio():
    input_file = fixtures / "biblio.json"
    output_file = fixtures / "biblio_output.json"
    input_data = json.loads(input_file.read_text())
    output_data = PublicSearchBiblioPage.model_validate(input_data)
    # output_file.write_text(output_data.model_dump_json(indent=2))
    expected_data = json.loads(output_file.read_text())
    compare_dicts(json.loads(output_data.model_dump_json()), expected_data)


def test_parse_doc():
    input_file = fixtures / "doc.json"
    output_file = fixtures / "doc_output.json"
    input_data = json.loads(input_file.read_text())
    output_data = [PublicSearchDocument.model_validate(o) for o in input_data]
    # output_file.write_text(json.dumps([o.model_dump() for o in output_data], indent=2, default=json_serial))
    expected_data = json.loads(output_file.read_text())
    output_data = json.loads(json.dumps([o.model_dump() for o in output_data], indent=2, default=json_serial))
    compare_lists(output_data, expected_data)



from copy import deepcopy
from typing import Callable
from typing import Generic
from typing import Optional

import inflection

from .api import PtabApi
from .model import PtabDecision
from .model import PtabDocument
from .model import PtabProceeding
from .util import peds_to_ptab
from patent_client.util.manager import Manager
from patent_client.util.manager import ModelType
from patent_client.util.request_util import get_start_and_row_count


class PtabManager(Manager, Generic[ModelType]):
    url = "https://developer.uspto.gov/ptab-api"
    page_size = 25
    instance_schema = None
    api_method: Optional[Callable] = None

    async def _aget_results(self):
        query = deepcopy(self.config.filter)
        query = peds_to_ptab(query)
        query["sort"] = " ".join(inflection.camelize(o, uppercase_first_letter=False) for o in self.config.order_by)
        for start, rows in get_start_and_row_count(self.config.limit, self.config.offset, self.page_size):
            query["start"] = start
            query["rows"] = rows
            page = await self.api_method(**query)
            for doc in page.docs:
                yield doc
            if len(page.docs) < rows:
                break

    async def alen(self):
        page = await self.api_method(**peds_to_ptab(self.config.filter))
        return min(self.config.limit, page.num_found) if self.config.limit else page.num_found

    def allowed_filters(self):
        params = schema_doc["paths"][self.path]["get"]["parameters"]
        return {inflection.underscore(p["name"]): p["description"] for p in params}


class PtabProceedingManager(PtabManager[PtabProceeding]):
    path = "/proceedings"
    default_filter = "proceeding_number"
    api_method = PtabApi.get_proceedings


class PtabDocumentManager(PtabManager[PtabDocument]):
    path = "/documents"
    default_filter = "document_identifier"
    api_method = PtabApi.get_documents


class PtabDecisionManager(PtabManager[PtabDecision]):
    path = "/decisions"
    default_filter = "identifier"
    api_method = PtabApi.get_decisions
